###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# https://microsoft.github.io/pyright/#/builtins
# This file is only used for the Python extension on VScode
import builtins

# We just create a new definition for Undefined
Undefined: builtins.Undefined # pylint: disable=redefined-builtin



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['JSONRequest', 'Request']
from starlette.requests import Request as _Request

from everysk.core.compress import decompress
from everysk.core.log import Logger
from everysk.core.serialize import loads


log = Logger(__name__)


class Request(_Request):

    async def body(self) -> bytes:
        """
        Method to return the body of the request.
        If we receive a request with a Content-Encoding header, we decompress the body.
        """
        # If self._body is set then body was already read.
        if not hasattr(self, '_body'):
            body = await super().body()
            # We only decompress the body if it is not empty.
            if body:
                content_encoding = self.headers.get('Content-Encoding', '').lower()
                if 'gzip' in content_encoding:
                    try:
                        body = decompress(body, protocol='gzip', serialize=None )
                    except Exception: # pylint: disable=broad-except
                        log.error('Error decompressing the request body.', extra={'labels': {'body': body}})

            self._body = body # pylint: disable=attribute-defined-outside-init
        return self._body


class JSONRequest(Request):

    async def json(self):
        """
        Method to return the JSON content of the request body.
        It uses the loads function from the core.serialize module to parse the body.
        """
        if not hasattr(self, '_json'):
            body = await self.body()
            self._json = loads(body, protocol='json', use_undefined=True) # pylint: disable=attribute-defined-outside-init

        return self._json



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import BoolField, ChoiceField, DictField, IntField, ListField, StrField

## Project settings
DEBUG = BoolField(default=True)

# PROD | DEV | LOCAL
PROFILE = ChoiceField(default='DEV', choices=('PROD', 'DEV', 'LOCAL'))

## Redis
REDIS_HOST = StrField(default='0.0.0.0')
REDIS_PORT = IntField(default=6379)

## Everysk SIGNING key is used to verify the integrity of the data
# Remember to change this key in production
EVERYSK_SIGNING_KEY = StrField(default='c90d39a0e065d56672393fc0102ac8e8680cc274744ce60fa269e6ab08d86673')

## Google Cloud
EVERYSK_GOOGLE_CLOUD_LOCATION = StrField(default='us-central1')
EVERYSK_GOOGLE_CLOUD_PROJECT = StrField()

## Logger settings
# This is the default app server to infer the headers and payload
LOGGING_APP_SERVER = StrField(default='flask')

# This is the default string to create the Google Cloud Trace ID
LOGGING_GOOGLE_CLOUD_TRACE_ID = StrField(default='projects/{EVERYSK_GOOGLE_CLOUD_PROJECT}/traces')

# This is the default Formatter for the logs
LOGGING_JSON = BoolField(default=False)

## Slack URL to send messages
SLACK_URL = StrField()

## Activate/deactivate HTTP Log for every request/response
EVERYSK_HTTP_LOG_RESPONSE = BoolField(default=False)

## List of HTTP Success status codes that do not raise HTTPError
HTTP_SUCCESS_STATUS_CODES = ListField(default=[200, 201, 202, 204, 303], readonly=True)

## HTTP default headers used to send requests
HTTP_DEFAULT_HEADERS = DictField(
    default={
        'Accept-Encoding': 'gzip',
        'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
        'Cache-control': 'no-cache',
        'Content-Type': 'text/html; charset=UTF-8',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
    },
    readonly=True
)
HTTP_DEFAULT_TIMEOUT = IntField(default=30)
HTTP_DEFAULT_SSL_VERIFY = BoolField(default=True)
HTTP_DEFAULT_RETRY_LIMIT = IntField(default=5)
HTTP_DEFAULT_RETRY_END_SECONDS = IntField(default=30)
HTTP_DEFAULT_RETRY_START_SECONDS = IntField(default=5)

## Serialization settings
SERIALIZE_CONVERT_METHOD_NAME = StrField(default='to_native')
SERIALIZE_DATE_KEY = StrField(default='__date__')
SERIALIZE_DATETIME_KEY = StrField('__datetime__')
SERIALIZE_UNDEFINED_KEY = StrField('__undefined__')
SERIALIZE_USE_UNDEFINED = BoolField(default=False)

## Activate/deactivate the use of the verify flag on HTTP requests.
# By default the value is Undefined and the value comes from the self._config.ssl_verify
# attribute in the HttpConnection class, if it is defined we use it.
HTTP_REQUESTS_VERIFY = BoolField(default=Undefined)

# Activate/deactivate the use of random user agents on HTTP requests
HTTP_USE_RANDOM_USER_AGENT = BoolField(default=False)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import re
from everysk.core.fields import BoolField, StrField


_EVERYSK_PRIVATE = StrField(default='private-attribute')
EVERYSK_TEST_NAME = StrField(default='test-case', readonly=True)
EVERYSK_TEST_FAKE = True
EVERYSK_TEST_INT: int = 1
EVERYSK_TEST_FIELD_INHERIT = StrField(default='{EVERYSK_TEST_NAME} as {EVERYSK_TEST_FAKE}')
EVERYSK_TEST_VAR_INHERIT = '{EVERYSK_TEST_FAKE} as {EVERYSK_TEST_NAME}'
EVERYSK_TEST_BOOL_VAR = BoolField(default=False)

# Settings with default values
EVERYSK_TEST_STR_DEFAULT_NONE: str = None
EVERYSK_TEST_STR_DEFAULT_UNDEFINED: str = Undefined

# https://everysk.atlassian.net/browse/COD-4197
EVERYSK_TEST_RE_PATTERN: re.Pattern = re.compile(r'[azAZ]')
EVERYSK_TEST_RE_PATTERN_NONE: re.Pattern = None
EVERYSK_TEST_RE_PATTERN_UNDEFINED: re.Pattern = Undefined



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from copy import copy, deepcopy
from everysk.core.undefined import UndefinedType
from everysk.core.unittests import TestCase


class UndefinedTestCase(TestCase):

    def test_sitecustomize_file(self):
        # If Undefined exists the src/sitecustomize.py is working
        Undefined # pylint: disable=pointless-statement

    def test_init(self):
        msg = 'Do not use this class, use the constant Undefined.'
        self.assertRaisesRegex(
            NotImplementedError,
            msg,
            UndefinedType
        )

    def test_bool(self):
        self.assertFalse(Undefined)

    def test_copy(self):
        self.assertEqual(copy(Undefined), Undefined)

    def test_deepcopy(self):
        self.assertEqual(deepcopy(Undefined), Undefined)

    def test_delattr(self):
        with self.assertRaisesRegex(AttributeError, UndefinedType.default_error_message):
            del Undefined._block

    def test_equal(self):
        new_undefined = Undefined
        copy_undefined = copy(Undefined)
        deepcopy_undefined = deepcopy(Undefined)
        self.assertEqual(new_undefined, Undefined)
        self.assertEqual(copy_undefined, Undefined)
        self.assertEqual(deepcopy_undefined, Undefined)
        self.assertTrue(new_undefined is Undefined)
        self.assertTrue(copy_undefined is Undefined)
        self.assertTrue(deepcopy_undefined is Undefined)

    def test_getattr(self):
        with self.assertRaisesRegex(AttributeError, UndefinedType.default_error_message):
            getattr(Undefined, 'banana')

    def test_repr(self):
        self.assertEqual(repr(Undefined), '<Undefined value>')

    def test_setattr(self):
        with self.assertRaisesRegex(AttributeError, UndefinedType.default_error_message):
            setattr(Undefined, 'banana', 'amarela')

    def test_str(self):
        self.assertEqual(str(Undefined), '<Undefined value>')

    def test_hash(self):
        hash(Undefined)
        self.assertSetEqual({Undefined}, {Undefined})



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources import WorkflowExecution
from everysk.core.unittests import TestCase, mock

###############################################################################
#   Workflow Execution TestCase Implementation
###############################################################################
class APIWorkflowExecutionTestCase(TestCase):

    def test_workflow_execution_class_name(self):
        self.assertEqual(WorkflowExecution.class_name(), 'workflow_execution')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_workflow_execution_retrieve_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()
        mock_create_api_requestor.return_value = mock_api_requestor

        workflow_id = '12345'
        mock_response_data = {'workflow_execution': {'key': 'data'}}
        mock_api_requestor.get.return_value = mock_response_data
        kwargs = {'param1': 'value1', 'param2': 'value2'}
        expected_url = f'/workflows/{workflow_id}/workflow_executions'
        result = WorkflowExecution.retrieve(workflow_id, **kwargs)

        mock_api_requestor.get.assert_called_once_with(expected_url, kwargs)
        self.assertIsInstance(result, WorkflowExecution)
        self.assertEqual(result['key'], 'data')

    @mock.patch('everysk.api.api_resources.worker_execution.WorkerExecution.retrieve')
    @mock.patch('everysk.api.api_resources.workflow_execution.WorkflowExecution.retrieve')
    @mock.patch('everysk.api.utils.sleep', return_value=None)
    def test_synchronous_retrieve_method(self, mock_sleep, mock_retrieve, mock_worker_retrieve):
        mock_retrieve.side_effect= [
            {'status': 'IN_PROGRESS', 'ender_worker_execution_id': 'worker1'},
            {'status': 'COMPLETED', 'ender_worker_execution_id': 'worker1'}
        ]
        worker_execution_result = {'worker_execution': 'result'}
        mock_worker_retrieve.return_value = worker_execution_result
        workflow_id = '12345'
        result = WorkflowExecution.syncronous_retrieve(workflow_id)

        self.assertEqual(mock_retrieve.call_count, 2)
        mock_retrieve.assert_any_call(workflow_id)
        mock_worker_retrieve.assert_called_once_with(worker_execution_id='worker1', with_result=True)
        self.assertEqual(result, worker_execution_result)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api import utils
from everysk.core.string import pluralize, snake_case

###############################################################################
#   APIResource Implementation
###############################################################################
class APIResource(utils.EveryskObject):
    """
    Represents a resource accessible via an API

    This class serves as a base class for API resources and provides method for refreshing resource data and obtaining class names and URLs.
    """
    def __init__(self, retrieve_params, params) -> None:
        super().__init__(retrieve_params, params)
        self.__retrieve_params = retrieve_params

    def refresh(self, **kwargs):
        """
        Refreshes the resource data from the API.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            APIResource: The refreshed APIResource object.
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f"{self.class_url()}/{self.get('id')}"
        response = api_req.get(url, self.__retrieve_params)
        data = response[self.class_name()]
        self.update(data)
        self.clear_unsaved_values()
        return self

    @classmethod
    def class_name(cls):
        """
        convert the class name to the snake case format
        """
        return snake_case(cls.__name__)

    @classmethod
    def class_name_list(cls):
        """
        Returns the pluralized class name.

        Returns:
            str: The pluralized class name.
        """
        return pluralize(cls.class_name())

    @classmethod
    def class_url(cls):
        """
        Returns the URL path for the class.
        Utilizing the class name list and ensuring consistency across different parts of the code.

        Returns:
            str: The URL path for the class.
        """
        return f'/{cls.class_name_list()}'

###############################################################################
#   Retrievable APIResource Implementation
###############################################################################
class RetrievableAPIResource(APIResource):

    @classmethod
    def retrieve(cls, id, **kwargs):
        """
        Retrieves a single resource from the API by its ID.
        Sends a GET request to the API to retrieve a single resource of the specified class type
        identified by the provided ID.

        Args:
            id (str):
                The unique identifier of the resource.

            workspace (str):
                Determines on which workspace the request will be made

            **kwargs (dict):
                Any additional keyword arguments to customize the request

        Returns:
            object: A single resource of the class type.
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'{cls.class_url()}/{id}'
        response = api_req.get(url, kwargs)
        return utils.to_object(cls, kwargs, response)

###############################################################################
#   Listable APIResource Implementation
###############################################################################
class ListableAPIResource(APIResource):

    @classmethod
    def list(cls, **kwargs):
        """
        Retrieves a list of resources from the API.
        This class method sends a GET request to the API to retrieve a list of resources of the specified class type.
        Check https://everysk.com/api/docs/#list-all-portfolios for additional information

        Args:
            query (str, optional):
                Request a list of portfolios fileting it by `name` or `tag`

            workspace (str, optional):
                Determines on which workspace the request will be made

            page_size (int, optional):
                Set the number of object that will be listed per page

            page_token (int, optional):
                This token defines which page will be returned to the user

            **kwargs (dict):
                Additional keyword arguments to customize the request

        Returns:
            list: A list of resource of the class type.
        """
        api_req = utils.create_api_requestor(kwargs)
        url = cls.class_url()
        response = api_req.get(url, kwargs)
        return utils.to_list(cls, kwargs, response)

    @classmethod
    def auto_paging_iter(cls, **kwargs):
        """
        Provides an iterator over all resources, handling pagination automatically.
        This class method iterates over all resources of the specified class type, handling pagination automatically
        to retrieve the complete list of resources. Check https://everysk.com/api/docs/#pagination for more information

        Args:
            page_size (int):
                A limit on the number of objects to be returned, between 1 and 100

            page_token (str):
                Returns the previous page token

            **kwargs (dict):
                Additional keyword arguments to customize the request, such as filters or pagination parameters.

        Returns:
            object: A resource of the class type.
        """
        params = dict(kwargs)
        page = cls.list(**params)
        while True:
            for item in page:
                yield item
            if page.next_page_token() is None:
                return
            params['page_token'] = page.next_page_token()
            page = cls.list(**params)

###############################################################################
#   Deletable APIResource Implementation
###############################################################################
class DeletableAPIResource(APIResource):

    def delete(self):
        """
        Deletes the current instance of the resource from the API
        This class method sends a DELETE request to the API to delete the current instance.

        Example:
            >>> from everysk.api.api_resources.api_resource import DeletableAPIResource
            >>> resource = DeletableAPIResource.retrieve("resource_id")
            >>> resource.delete()
        """
        api_req = utils.create_api_requestor()
        url = f"{self.class_url()}/{self.get('id')}"
        workspace = self.get('workspace', None)
        if workspace:
            url = f'{url}?workspace={workspace}'
        response = api_req.delete(url)
        data = response[self.class_name()]
        self.clear()
        self.update(data)
        self.clear_unsaved_values()
        return self

    @classmethod
    def remove(cls, id, **kwargs):
        """
        Deletes a resource from the API by its ID.
        Sends a DELETE request to the API to delete a resource of the specified class type
        identified by the provided ID.

        Args:
            id: The unique identifier of the resource to be deleted.

            workspace (str, optional):
                Determines on which workspace the request will be made.

            **kwargs (dict): Additional keyword arguments to customize the request.

        Example:
            >>> DeletableAPIResource.remove('resource_id')
            Deleted resource_object
        """
        api_req = utils.create_api_requestor()
        url = f'{cls.class_url()}/{id}'
        workspace = kwargs.get('workspace', None)
        if workspace:
            url = f'{url}?workspace={workspace}'
        response = api_req.delete(url)
        data = response[cls.class_name()]
        return utils.to_object(cls, {}, response)

###############################################################################
#   Creatable APIResource Implementation
###############################################################################
class CreateableAPIResource(APIResource):

    @classmethod
    def create(cls, **kwargs):
        """
        Creates a new instance of the resource and saves it to the API.
        This class method sends a POST request to the API to create a new instance of the resource with the provided data.

        Args:
            description (str, optional):
                Provides detailed information about the resource

            tags (list, optional):
                Sequence of hashtags used to label the related resources. Any sequence of lowercase character, numbers, and underscore might be used.

            date (str, optional):
                Resource date in the following format: `YYYYMMDD`

            workspace (str, optional):
                Determines on which workspace the request will be made

            **kwargs (dict):
                Keyword arguments representing the data for creating the resource. Depending on the resource, each entity will have different kwargs.

        Example:
            >>> new_resource_data = {'name': 'New Resource'}
            >>> new_resource = CreatableAPIResource.create(**new_resource_data)
            >>> print(new_resource)
            New resource_object
        """
        api_req = utils.create_api_requestor(kwargs)
        url = cls.class_url()
        response = api_req.post(url, kwargs)
        return utils.to_object(cls, kwargs, response)

###############################################################################
#   Updatable APIResource Implementation
###############################################################################
class UpdateableAPIResource(APIResource):

    @classmethod
    def modify(cls, id, **kwargs):
        """
        Modifies a resource by its ID with the provided data.

        This class method sends a PUT request to the API to modify a resource of the specified class type
        identified by the provided ID with the provided data.

        Args:
            id (str):
                The unique identifier of the resource to be modified.

            **kwargs (dict):
                Keyword arguments representing the data to be updated. The kwargs will differ for each entity.

        Example:
            >>> updated_data = {'name': 'Updated Resource'}
            >>> updated_resource = UpdateableAPIResource.modify('resource_id', **updated_data)
            >>> print(updated_resource)
            Updated resource_object
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'{cls.class_url()}/{id}'
        response = api_req.put(url, kwargs)
        data = response[cls.class_name()]
        return utils.to_object(cls, kwargs, response)

    def save(self, **kwargs):
        """
        Saves the changes made to the current instance of the resource to the API.
        This method sends a PUT request to the API to save the changes made to the current instance of the resource.

        Args:
            **kwargs (dict):
                each entity will have different arguments for saving

        Example:
            >>> resource = UpdateableAPIResource.retrieve('resource_id')
            >>> resource['name'] = 'Updated Name'
            >>> resource.save()
            Updated resource_object
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f"{self.class_url()}/{self.get('id')}"
        #response = api_req.put(url, self)
        unsaved_values = self.get_unsaved_values()
        response = api_req.put(url, unsaved_values)
        data = response[self.class_name()]
        self.update(data)
        self.clear_unsaved_values()
        return self

###############################################################################
#   Filterable APIResource Implementation
###############################################################################
class FilterableAPIResource(APIResource):

    @classmethod
    def filter(cls,**kwargs):
        """
        Filters resources based on the provided criteria.

        This class method sends a POST request to the API to filter resources of the specified class type
        based on the provided criteria.

        Args:
            limit (int, optional): The maximum number of resources to retrieve. Defaults to a predetermined limit if not specified.

            tags (list[str], optional):
                A list of strings representing tags used to filter resources. Tags should be a sequence
                of lowercase characters, numbers, and underscores.

            start (str, optional):
                A start value used for filtering. The format and usage of this parameter can vary
                depending on the resource type and should be a string or null.

            end (str, optional):
                An end value for filtering, similar in format and function to `start`. It should be a string or null.

            date_time (str, optional):
                A datetime string used for filtering resources. The expected format should be compliant
                with the API's datetime format requirements.

            link_uid (str, optional):
                A unique identifier used to filter resources by their association to another entity.

            **kwargs (dict):
                Additional keyword arguments that represent other filtering criteria not explicitly listed. These should
                align with the filterable fields of the resources.

        Returns:
            list: A list of filtered resources.

        Example:
            >>> filtered_resources = FilterableAPIResource.filter(param1=value1, param2=value2)
            >>> print(filtered_resources)
            [filtered_resource1, filtered_resource2, ...]
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'/{cls.class_name_list()}/filter'
        response = api_req.post(url, kwargs)
        return utils.to_list(cls, kwargs, response)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from everysk.core.exceptions import FieldValueError, ReadonlyError, RequiredError
from everysk.core import slack
from everysk.core.unittests import TestCase, mock


class SlackTestCase(TestCase):

    def test_is_json(self):
        message = "The field 'is_json' value cannot be changed."
        self.assertRaisesRegex(
            FieldValueError,
            message,
            slack.Slack,
            title='Test',
            message='Message',
            is_json=False
        )
        with self.assertRaisesRegex(FieldValueError, message):
            obj = slack.Slack(title='Test', message='Message')
            obj.is_json = False

    def test_title(self):
        message = 'The title attribute is required.'
        self.assertRaisesRegex(RequiredError, message, slack.Slack, message='Message')

    def test_message(self):
        message = 'The message attribute is required.'
        self.assertRaisesRegex(RequiredError, message, slack.Slack, title='Title')

    def test_choices(self):
        message = r"The value 'banana' for field 'color' must be in this list \(None, 'danger', 'success', 'warning'\)."
        self.assertRaisesRegex(
            FieldValueError,
            message,
            slack.Slack,
            title='Test',
            message='Message',
            color='banana'
        )
        with self.assertRaisesRegex(FieldValueError, message):
            obj = slack.Slack(title='Test', message='Message')
            obj.color = 'banana'

    def test_color_map(self):
        message = "The field '_color_map' value cannot be changed."
        self.assertRaisesRegex(
            FieldValueError,
            message,
            slack.Slack,
            title='Test',
            message='Message',
            _color_map={'banana': '#f3f3f3'}
        )
        with self.assertRaisesRegex(FieldValueError, message):
            obj = slack.Slack(title='Test', message='Message')
            obj._color_map = {'banana': '#f3f3f3'}

        with self.assertRaisesRegex(ReadonlyError, 'This field value cannot be changed.'):
            obj = slack.Slack(title='Test', message='Message')
            obj._color_map['banana'] = '#f3f3f3'

    def test_get_payload(self):
        obj = slack.Slack(title='Test', message='Message')
        self.assertDictEqual(
            obj.get_payload(),
            {
                'attachments': [{
                    'color': '#000000',
                    'blocks': [
                        {
                            'text': {'emoji': True, 'text': 'Test', 'type': 'plain_text'},
                            'type': 'header'
                        },
                        {'type': 'divider'},
                        {
                            'text': {'text': 'Message', 'type': 'mrkdwn'},
                            'type': 'section'
                        },
                        {'type': 'divider'}
                    ],
                }]
            }
        )

    @mock.patch.object(slack.Slack, 'get_response')
    def test_send(self, get_response: mock.MagicMock):
        slack.Slack(title='Test', message='Message').send()
        get_response.assert_called_once_with()



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# Classes that handle HTTP Connections.

# Url: https://requests.readthedocs.io/en/latest/
#      https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers
#      https://developer.mozilla.org/en-US/docs/Glossary/Quality_values
#
import json
import time
from os.path import dirname
from random import SystemRandom
from types import TracebackType
from typing import Any, Self

import httpx
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.exceptions import HttpError, InvalidArgumentError
from everysk.core.fields import BoolField, DictField, IntField, ListField, StrField
from everysk.core.log import Logger, _get_gcp_headers
from everysk.core.object import BaseObject, BaseObjectConfig
from everysk.core.serialize import loads


log = Logger(name='everysk-lib-core-http-log')


class HttpConnectionConfig(BaseObjectConfig):
    # Number of connections to keep opened in the pool.
    pool_connections_min = IntField(default=20, min_size=1)

    # Maximum number of connections in the pool.
    pool_connections_max = IntField(default=100)

    # Timeout to wait before closing idle connections.
    pool_connections_timeout = IntField(default=5)

    # If the connection should follow the redirects automatic or raise HTTP Redirect error.
    follow_redirects = BoolField(default=True)

    # If the connection accepts HTTP 1 protocol.
    use_http1 = BoolField(default=True)

    # If the connection accepts HTTP 2 protocol.
    use_http2 = BoolField(default=True)

    # Keep the full path to the file with the user agents.
    user_agents_file = StrField()

    # List of user agents to be used in the requests.
    user_agents = ListField()

    # Activate/deactivate the use of the verify flag on HTTP requests.
    # By default this is defined in settings.HTTP_REQUESTS_VERIFY
    # but could be defined in the class configuration too.
    ssl_verify = BoolField(default=settings.HTTP_DEFAULT_SSL_VERIFY)

    # Limit for retries
    retry_limit = IntField(default=settings.HTTP_DEFAULT_RETRY_LIMIT)

    # Times that randrange will use to do the next retry
    retry_end_seconds = IntField(default=settings.HTTP_DEFAULT_RETRY_END_SECONDS)
    retry_start_seconds = IntField(default=settings.HTTP_DEFAULT_RETRY_START_SECONDS)

    def __after_init__(self) -> None:
        # Load the user agents from the file
        if self.user_agents_file is None:
            base_dir = dirname(__file__)
            self.user_agents_file = f'{base_dir}/fixtures/user_agents.json'

        if self.user_agents is None:
            with open(self.user_agents_file, 'r', encoding='utf-8') as fd:
                self.user_agents = json.load(fd)

    def get_client(self) -> httpx.Client:
        """
        Returns a httpx.Client object with the configurations set in the class.
        """
        limits = httpx.Limits(
            max_keepalive_connections=self.pool_connections_min,
            max_connections=self.pool_connections_max,
            keepalive_expiry=self.pool_connections_timeout
        )
        return httpx.Client(
            follow_redirects=self.follow_redirects,
            limits=limits,
            http1=self.use_http1,
            http2=self.use_http2,
            verify=self.get_ssl_verify()
        )

    def get_ssl_verify(self) -> bool:
        """
        Returns a boolean indicating whether the SSL value
        is set.

        Returns:
            bool: True if the SSL verify is set, False otherwise.

        Example:
            >>> http_connection_config = HttpConnectionConfig()
            >>> http_connection_config.get_ssl_verify()
            True
        """
        return self.ssl_verify if settings.HTTP_REQUESTS_VERIFY is Undefined else settings.HTTP_REQUESTS_VERIFY

    def get_random_agent(self) -> str:
        """
        Return a random user agent from the list of user agents.
        """
        random = SystemRandom()
        return random.choice(self.user_agents)


###############################################################################
#   HttpConnection Class Implementation
###############################################################################
class HttpConnection(BaseObject):
    """
    Base class to use for HTTP connections, has two attributes:
        - timeout: It's in and represent seconds, defaults to 30.
        - url: It's string and will be the destination.
    """
    class Config(HttpConnectionConfig):
        pass

    ## Private attributes
    _client: httpx.Client = None
    _config: Config = None # To autocomplete correctly
    _retry_count = IntField(default=1) # Used to control how many times this connection was retry

    ## Public attributes
    headers = DictField(default=None)
    timeout = IntField(default=settings.HTTP_DEFAULT_TIMEOUT) # This is read timeout
    url = StrField(default=None)

    ## Private methods
    def __call__(self, **kwargs: dict) -> Self:
        """
        Set the attributes of the class with the values passed as kwargs.
        We use this method together with the with statement to set the values of the class.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

        return self

    def __enter__(self) -> Self:
        """
        Create and set the HTTP client to be used in the connection so we could reuse the connection.
        """
        self._client = self._config.get_client()
        return self

    def __exit__(self, __exc_type: type[BaseException] | None, __exc_value: BaseException | None, __traceback: TracebackType | None) -> bool | None:
        """
        Close the HTTP client to free the resources.

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        self._client.close()
        self._client = None
        return False

    def _clean_response(self, response: httpx.Response) -> httpx.Response:
        """
        Checks status_code for response, if status_code is different than 200 throws an exception.

        Args:
            response (httpx.Response): Http response from server.

        Raises:
            HttpError: If something goes wrong raise exception with status_code and content.
        """
        if getattr(response, 'status_code', settings.HTTP_SUCCESS_STATUS_CODES[0]) not in settings.HTTP_SUCCESS_STATUS_CODES:
            raise HttpError(status_code=response.status_code, msg=response.content)

        return response

    def _get_headers(self) -> dict:
        try:
            # We try to get the GCP headers to send the request
            # The first attempt is to get from a context var
            # If it fails we try to get from the server running
            gcp_headers = _get_gcp_headers()
        except Exception: # pylint: disable=broad-exception-caught
            gcp_headers = {}

        # Get the headers from the class or child classes
        headers = self.get_headers()

        # Update GCP headers with the headers from the class
        # so if the class has the same key it will be overwritten
        gcp_headers.update(headers)

        return gcp_headers

    def _get_response_from_url(self) -> httpx.Response:
        """
        This will be implemented on child classes to really do the connection.
        """
        return None

    ## Public methods
    def get_headers(self) -> dict:
        """
        Headers needed to send HTTP methods.
        Below are the most common Headers used by browsers,
        we use them to look less like a Bot and more like a valid access.

        Returns:
            dict: A dictionary containing the headers information.

        Example:
            >>> http_connection = HttpConnection()
            >>> http_connection.get_headers()
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'text/html; charset=UTF-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        """
        headers = settings.HTTP_DEFAULT_HEADERS.copy()
        if settings.HTTP_USE_RANDOM_USER_AGENT:
            headers['User-Agent'] = self._config.get_random_agent()

        if self.headers is not None:
            headers.update(self.headers)

        return headers

    def get_url(self) -> str:
        """
        Generate the correct url to fetch data from vendor on POST/GET requests.

        Returns:
            str: Containing the correct URL.
        """
        return self.url

    def get_timeout(self) -> httpx.Timeout:
        """
        Return the timeout to be used in the connection.
        """
        timeout = httpx.Timeout(
            timeout=30, # Default timeout for all operations
            read=self.timeout # Timeout for reading the response
        )
        return timeout

    def message_error_check(self, message: str, status_code: int) -> bool: # pylint: disable=unused-argument
        """
        If this method returns True, the connection will be tried again.

        Args:
            message (str): The error message that occurred on the connection.
            status_code (int): The status code of the response.
        """
        return False

    def get_response(self) -> httpx.Response:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again for a few more times.
        """
        try:
            response = self._clean_response(self._get_response_from_url())
            # After a success we set the value to 1 again
            self._retry_count = 1
        except Exception as error: # pylint: disable=broad-exception-caught
            # Sometimes it can happen that the server is busy, if this happen the error message must be tested
            # and must return true to enable recursion and we will try again the connection.
            message = str(error).lower()
            status_code = getattr(error, 'status_code', 500)
            if self.message_error_check(message, status_code) and self._retry_count < self._config.retry_limit:
                self._retry_count += 1
                # As we have several processes, we use a random number to avoid collision between them.
                random = SystemRandom()
                time.sleep(random.randint(self._config.retry_start_seconds, self._config.retry_end_seconds))
                response = self.get_response()
            else:
                raise error

        return response


class HttpGETConnection(HttpConnection):
    """ Class that implements a interface for HTTP GET connections """
    params = DictField()
    user = StrField()
    password = StrField()

    def __init__(self, url: str = None, headers: dict = None, params: dict = None, timeout: int = None, user: str = None, password: str = None, **kwargs) -> None:
        """
        HTTP GET Connection used to fetch data from a URL.
        Url could be passed on the constructor, set in the class or set later.

        Args:
            url (str, optional): The URL to fetch data from.
            headers (dict, optional): Extra headers added in the connection. Defaults to None.
            params (dict, optional): Parameters added in the url as query. Defaults to None.
            timeout (int, optional): Time when the connection stops to wait and raise an Exception. Defaults to 30 seconds.
            user (str, optional): Username if needed for authentication. Defaults to None.
            password (str, optional): _description_. Defaults to None.
        """
        # We only sent to the super the values that are not None
        # to not overwrite the default values or the ones that are set in the class.
        if url is not None:
            kwargs['url'] = url
        if headers is not None:
            kwargs['headers'] = headers
        if params is not None:
            kwargs['params'] = params
        if timeout is not None:
            kwargs['timeout'] = timeout
        if user is not None:
            kwargs['user'] = user
        if password is not None:
            kwargs['password'] = password

        super().__init__(**kwargs)

    def get_params(self) -> dict:
        """
        This method is used to make the correct params to pass on GET request.
        These params will be added to the URL with & separating then.
        """
        return self.params

    def _get_response_from_url(self) -> httpx.Response:
        """
        Try to fetch data from url using GET request.
        Note that any dictionary key whose value is None will not be added to the URL's query string.
        """
        params = {
            'url': self.get_url(),
            'headers': self._get_headers(),
            'params': self.get_params(),
            'timeout': self.get_timeout()
        }
        if self.user:
            params['auth'] = (self.user, self.password)

        if settings.EVERYSK_HTTP_LOG_RESPONSE:
            dct = params.copy()
            # To remove the password in the logs
            if 'auth' in params:
                dct['auth'] = (params['auth'][0], '***********')

            log.debug('HTTP GET request: %s', self.get_url(), extra={'labels': dct})

        # If the client is already set and not closed we use it
        # otherwise we create a new client and close it after the request.
        if self._client and not self._client.is_closed:
            response = self._client.get(**params)
        else:
            with self._config.get_client() as client:
                response = client.get(**params)

        if settings.EVERYSK_HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP GET response: %s', self.get_url(), extra={'labels': dct})

        return response


class HttpPOSTConnection(HttpConnection):
    """
    Class that implements a interface for HTTP POST connections.
    If self.is_json is True the POST method will be a JSON POST,
    otherwise will be a Form POST Data.
    """
    is_json = BoolField(default=True)
    payload = DictField()

    def __init__(self, url: str = None, headers: dict = None, payload: dict = None, timeout: int = None, is_json: bool = None, **kwargs) -> None:
        """
        HTTP POST Connection used to send data to a URL.
        Url could be passed on the constructor, set in the class or set later.

        Args:
            url (str, optional): The URL to send data.
            headers (dict, optional): Extra headers added in the connection. Defaults to None.
            payload (dict, optional): Data to send in the POST request. Defaults to None.
            timeout (int, optional): Time when the connection stops to wait and raise an Exception. Defaults to 30 seconds.
            is_json (bool, optional): If the POST method will be a JSON POST or a Form POST Data. Defaults to True.
        """
        # We only sent to the super the values that are not None
        # to not overwrite the default values or the ones that are set in the class.
        if url is not None:
            kwargs['url'] = url
        if headers is not None:
            kwargs['headers'] = headers
        if payload is not None:
            kwargs['payload'] = payload
        if timeout is not None:
            kwargs['timeout'] = timeout
        if is_json is not None:
            kwargs['is_json'] = is_json

        super().__init__(**kwargs)

    def get_headers(self) -> dict:
        """
        Headers needed to send HTTP Post methods.

        Returns:
            dict: The headers info for POST requests.

        Example:
            >>> http_post = HttpPostConnection()
            >>> http_post.get_headers()
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'application/json; charset=utf-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        """
        headers = super().get_headers()
        if self.is_json:
            headers['Content-Type'] = 'application/json; charset=utf-8'
        else:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        return headers

    def get_payload(self) -> dict:
        """
        Make the correct payload body to pass on POST request.

        Returns:
            dict: With all the payload information to send alongside the request.
        """
        return self.payload

    def _get_response_from_url(self) -> httpx.Response:
        """ Try to get/set data on url using POST request. """
        params = {
            'url': self.get_url(),
            'headers': self._get_headers(),
            'timeout': self.get_timeout()
        }
        if self.is_json:
            params['json'] = self.get_payload()
        else:
            params['data'] = self.get_payload()

        if settings.EVERYSK_HTTP_LOG_RESPONSE:
            log.debug('HTTP POST request: %s', self.get_url(), extra={'labels': params})

        # If the client is already set and not closed we use it
        # otherwise we create a new client and close it after the request.
        if self._client and not self._client.is_closed:
            response = self._client.post(**params)
        else:
            with self._config.get_client() as client:
                response = client.post(**params)

        if settings.EVERYSK_HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP POST response: %s', self.get_url(), extra={'labels': dct})

        return response


class HttpPOSTCompressedConnection(HttpPOSTConnection):

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        headers['Content-Encoding'] = 'gzip'
        return headers

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        return compress(self.payload, protocol='gzip', serialize='json', use_undefined=True, add_class_path=True)


class HttpSDKPOSTConnection(HttpPOSTCompressedConnection):

    is_json = BoolField(default=False, readonly=True)

    class_name = StrField()
    method_name = StrField()
    self_obj: Any = None
    params = DictField()
    timeout = IntField(default=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT)

    def get_url(self) -> str:
        return f'{settings.EVERYSK_SDK_URL}/{settings.EVERYSK_SDK_VERSION}/{settings.EVERYSK_SDK_ROUTE}'

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        self.payload = {
            'class_name': self.class_name,
            'method_name': self.method_name,
            'self_obj': self.self_obj,
            'params': self.params
        }
        return super().get_payload()

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        everysk_api_sid = settings.EVERYSK_API_SID
        everysk_api_token = settings.EVERYSK_API_TOKEN

        if not everysk_api_sid:
            raise InvalidArgumentError('Invalid API SID.')
        if not everysk_api_token:
            raise InvalidArgumentError('Invalid API TOKEN.')

        headers['Authorization'] = f'Bearer {everysk_api_sid}:{everysk_api_token}'

        return headers

    def get_response(self) -> dict:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again more 5 times.
        Decompress the response.content
        """
        response = super().get_response()
        return loads(response.content, use_undefined=True, instantiate_object=True)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import traceback
from typing import Any
from uuid import uuid4
from google.api_core.exceptions import AlreadyExists, DeadlineExceeded, ServiceUnavailable, TooManyRequests
from google.cloud.tasks_v2 import (
    CloudTasksClient, CreateTaskRequest, HttpMethod, PauseQueueRequest, ResumeQueueRequest, Task, Queue
)
from google.protobuf import duration_pb2
from everysk.core.compress import compress, decompress
from everysk.config import settings
from everysk.core.fields import IntField, StrField
from everysk.core.log import Logger
from everysk.core.object import BaseObject
from everysk.core.string import import_from_string


## Create Log for messages
log = Logger('everysk-workers')

###############################################################################
#   BaseGoogle Class Implementation
###############################################################################
class BaseGoogle(BaseObject):
    ## Private attributes
    _gtask: CloudTasksClient = None

    ## Public attributes
    google_task_project = StrField(default=settings.EVERYSK_GOOGLE_CLOUD_PROJECT)
    google_task_location = StrField(default=settings.EVERYSK_GOOGLE_CLOUD_LOCATION)
    worker_id = StrField(required=True)

    @property
    def gtask(self) -> CloudTasksClient:
        # pylint: disable=protected-access
        if self.__class__._gtask is None:
            self.__class__._gtask = CloudTasksClient()

        return self.__class__._gtask

###############################################################################
#   TaskGoogle Class Implementation
###############################################################################
class TaskGoogle(BaseGoogle):
    ## Public attributes
    google_task_name = StrField() # Must be a unique name if None is used, auto create this
    timeout = IntField(default=600, min_size=15, max_size=1800) # https://cloud.google.com/tasks/docs/reference/rest/v2/projects.locations.queues.tasks
    worker_url = StrField(required=True)

    ## Private methods
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        # This facilitate to init it over pickle
        self._received_kwargs = kwargs
        if self.google_task_name is None:
            self.google_task_name = f'{self.__class__.__name__}-{uuid4()}'

    ## GTask methods
    def gtask_queue_path(self) -> str:
        return self.gtask.queue_path(
            project=self.google_task_project,
            location=self.google_task_location,
            queue=self.worker_id
        )

    def gtask_task_path(self) -> str:
        return self.gtask.task_path(
            project=self.google_task_project,
            location=self.google_task_location,
            queue=self.worker_id,
            task=self.google_task_name
        )

    def gtask_create_task_request(self, body: dict) -> CreateTaskRequest:
        # https://cloud.google.com/tasks/docs/creating-http-target-tasks#createtask_method
        if self.timeout is not None:
            deadline = duration_pb2.Duration() # pylint: disable=no-member
            deadline.FromSeconds(self.timeout)

        task = {
            'name': self.gtask_task_path(),
            'dispatch_deadline': deadline,
            'http_request': {
                'http_method': HttpMethod.POST,
                'url': f'{self.worker_url}/worker/{self.worker_id}',
                'headers': self.get_headers(),
                'body': compress(body, serialize='pickle')
            }
        }
        return CreateTaskRequest(parent=self.gtask_queue_path(), task=task)

    ## Public methods
    def get_headers(self) -> dict:
        """
        Return the headers that must be used to receive the task.
        Remember that body must be always a byte object, then on child
        classes do like de example bellow.

        Example:
            >>> def get_headers(self) -> dict:
            ...     headers = super().get_headers()
            ...     headers['new_header'] = 'bla bla'
            ...     return headers
        """
        return {'Content-type': 'application/octet-stream'}

    def run(self) -> str:
        """ This must return a str object to be used on worker logs """
        pass

    def send_start(self) -> Queue:
        resume_queue_request = ResumeQueueRequest(name=self.gtask_queue_path())
        return self.gtask.resume_queue(request=resume_queue_request)

    def send_pause(self) -> Queue:
        pause_queue_request = PauseQueueRequest(name=self.gtask_queue_path())
        return self.gtask.pause_queue(request=pause_queue_request)

    def save(self, timeout: float = 30.0, retry_times: int = 0) -> Task:
        """
        Saves this task on Google Cloud Tasks to be executed.
        We will only retry if the Task Error is in (DeadlineExceeded, ServiceUnavailable, TooManyRequests).
        Be careful, if the task name is random retry could duplicate the task.
        # https://everysk.atlassian.net/browse/COD-1546

        Args:
            timeout (float, optional): Time to wait until the task is done. Defaults to 30.0.
            retry_times (int, optional): Retry times if the task fails. Defaults to 0.
        """
        # Timeout could not be greater than 30 seconds
        timeout = min(timeout, 30.0)

        body = {'cls': self.get_full_doted_class_path(), 'kwargs': self._received_kwargs}
        task = None
        try:
            task = self.gtask.create_task(request=self.gtask_create_task_request(body), timeout=timeout)

        except AlreadyExists:
            log.debug('Google task already exists: %s', self.gtask_task_path())

        except (DeadlineExceeded, ServiceUnavailable, TooManyRequests) as error:
            if retry_times > 0:
                retry_times = retry_times - 1
                task = self.save(timeout=timeout, retry_times=retry_times)
            else:
                raise error

        return task

###############################################################################
#   WorkerGoogle Class Implementation
###############################################################################
class WorkerGoogle(BaseGoogle):
    """
    Example:
        from flask import Flask, request
        from everysk.core.fields import StrField
        from everysk.core.workers import WorkerGoogle

        WORKER_ID = 'worker-id'

        class FirestoreSaveWorker(WorkerGoogle):
            google_task_location: StrField(default='location', readonly=True)
            google_task_project: StrField(default='project', readonly=True)
            worker_id: StrField(default=WORKER_ID, readonly=True)

        ## Flask app
        app = Flask(__name__)

        @app.route(f'/worker/{WORKER_ID}', methods=['POST'])
        def firestore_save():
            return FirestoreSaveWorker.worker_run(
                headers=request.headers,
                data=request.data
            )

        ## Main run function
        if __name__ == "__main__":
            app.run(host=settings.FLASK_HOST, port=settings.FLASK_PORT, debug=settings.DEBUG)
    """

    ## Public methods
    def check_google_task(self, headers: dict) -> bool:
        """ Check Google Task readers to ensure that the request come from then. """
        queue_name = headers.get('X-Cloudtasks-Queuename', None)
        user_agent = headers.get('User-Agent', None)
        return queue_name == self.worker_id and user_agent == 'Google-Cloud-Tasks'

    def run_task(self, cls: str, kwargs: dict) -> Any:
        """
        Run the task, if pause is activated from the Tread, stops
        until the self.worker_start message arrives.
        """
        cls = import_from_string(cls)
        task = cls(**kwargs)
        return task.run()

    @classmethod
    def worker_run(cls, headers: dict, data: bytes, worker_id: str = None) -> None:
        if worker_id is None:
            worker = cls()
        else:
            worker = cls(worker_id=worker_id)

        if not worker.check_google_task(headers):
            message = {'error': True, 'message': f"Couldn't validate Google headers - {headers}"}
            log.error(message['message'])
            return message

        try:
            post_data = decompress(data, serialize='pickle')
            result = worker.run_task(**post_data)
            return {'error': False, 'result': result}

        except Exception as error: #pylint: disable=broad-exception-caught
            message = f'Worker {cls.worker_id} error: {str(error)}'
            log.error('Worker %s error: %s', cls.worker_id, traceback.format_exc())
            return {'error': True, 'message': message}



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.datetime.datetime import DateTime, timezone, ZoneInfo
from everysk.core.datetime.date import Date
from everysk.core.datetime.date_mixin import timedelta



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.engines.cache import UserCache


###############################################################################
#   Cache Test Case Implementation
###############################################################################
class CacheTestCase(TestCase):
    def setUp(self) -> None:
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        self.expected_error_message = 'Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.'
        return super().setUp()

    ###############################################################################
    #   Get Method Test Case Implementation
    ###############################################################################
    def test_get_method_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'get', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'key': 'key', 'prefix': ''}}, protocol='gzip', serialize='json')
        expected_result = {}

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.get("key")

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    ###############################################################################
    #   Get Multi Method Test Case Implementation
    ###############################################################################
    def test_get_multi_method_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'get_multi', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'keys': ['key1', 'key2'], 'prefix': ''}}, protocol='gzip', serialize='json')
        expected_result = {}

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.get_multi(['key1', 'key2'])

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    ###############################################################################
    #   Set Method Test Case Implementation
    ###############################################################################
    def test_set_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'set', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'key': 'key', 'data': 'value', 'time': 14400, 'prefix': ''}}, protocol='gzip', serialize='json')
        expected_result = {}

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.set(key="key", data="value")

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    def test_set_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time=-2)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Set Multi Method Test Case Implementation
    ###############################################################################
    def test_set_multi_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'set_multi', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'data_dict': {'key1': 'value1', 'key2': 'value2'}, 'time': 14400, 'prefix': ''}}, protocol='gzip', serialize='json')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.set_multi({'key1': 'value1', 'key2': 'value2'})

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_set_multi_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_multi_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_multi_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Incr Method Test Case Implementation
    ###############################################################################
    def test_incr_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'incr', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'key': 'key', 'delta': 1, 'initial_value': 0, 'time': 14400}}, protocol='gzip', serialize='json')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.incr(key='key', delta=1, initial_value=0)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_incr_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_incr_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_incr_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Decr Method Test Case Implementation
    ###############################################################################
    def test_decr_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'decr', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'key': 'key', 'delta': 1, 'initial_value': 0, 'time': 14400}}, protocol='gzip', serialize='json')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.decr(key='key', delta=1, initial_value=0)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_decr_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_decr_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_decr_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Delete Method Test Case Implementation
    ###############################################################################
    def test_delete_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'delete', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'key': 'key'}}, protocol='gzip', serialize='json')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.delete(key='key')

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    ###############################################################################
    #   Delete Multi Method Test Case Implementation
    ###############################################################################
    def test_delete_multi_assert_is_called(self):
        cache = UserCache()
        expected_data = compress({'class_name': 'UserCache', 'method_name': 'delete_multi', 'self_obj': cache.to_dict(add_class_path=True), 'params': {'keys': ['key1', 'key2']}}, protocol='gzip', serialize='json')
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.delete_multi(keys=['key1', 'key2'])

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.datetime import DateTime
from everysk.core.exceptions import FieldValueError, RequiredError
from everysk.core.object import BaseDict
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.fields import EntityTagsField, CurrencyField, EntityNameField, EntityDescriptionField, EntityLinkUIDField, EntityWorkspaceField, EntityDateTimeField
from everysk.sdk.entities.tags import Tags


###############################################################################
#   Fields TestCase Implementation
###############################################################################
class _EntityNameTestAuxClass(BaseDict):
    required = EntityNameField()

class _EntityDescriptionFieldTestAuxClass(BaseDict):
    required = EntityDescriptionField()

class _EntityLinkUIDFieldTestAuxClass(BaseDict):
    required = EntityLinkUIDField()


class _EntityWorkspaceFieldTestAuxClass(BaseDict):
    required = EntityWorkspaceField()

class _EntityDateTimeFieldTestAuxClass(BaseDict):
    required = EntityDateTimeField()

class TestBaseCurrencyField(TestCase):

    def setUp(self):
        self.currency_field = CurrencyField(default="USD")

    def test_default_currencies(self):
        # Ensure the field was initialized with the correct set of default currencies
        self.assertIn(None, self.currency_field.choices)
        self.assertIn('USD', self.currency_field.choices)
        self.assertIn('BRL', self.currency_field.choices)

class TestEntityNameField(TestCase):

    def test_max_size(self):
        field = EntityNameField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '{settings.ENTITY_NAME_MAX_LENGTH + 1}' for attribute 'banana' must be between '{settings.ENTITY_NAME_MIN_LENGTH}' and '{settings.ENTITY_NAME_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value='a' * (settings.ENTITY_NAME_MAX_LENGTH + 1)
        )

    def test_min_size(self):
        field = EntityNameField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '0' for attribute 'banana' must be between '{settings.ENTITY_NAME_MIN_LENGTH}' and '{settings.ENTITY_NAME_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value=''
        )

    def test_required_lazy(self):
        cls = _EntityNameTestAuxClass()
        self.assertRaisesRegex(
            RequiredError,
            "The required attribute is required.",
            cls.validate_required_fields,
        )

    def test_empty_is_none(self):
        cls = _EntityNameTestAuxClass(required='')
        self.assertIsNone(cls.required)

class TestEntityDescriptionField(TestCase):

    def test_max_size(self):
        field = EntityDescriptionField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '{settings.ENTITY_DESCRIPTION_MAX_LEN + 1}' for attribute 'banana' must be between '{settings.ENTITY_DESCRIPTION_MIN_LEN}' and '{settings.ENTITY_DESCRIPTION_MAX_LEN}'.",
            field.validate,
            attr_name='banana',
            value='a' * (settings.ENTITY_DESCRIPTION_MAX_LEN + 1)
        )

    def test_empty_is_none(self):
        cls = _EntityDescriptionFieldTestAuxClass(required='')
        self.assertIsNotNone(cls.required)

class TestEntityLinkUIDField(TestCase):

    def test_max_size(self):
        field = EntityLinkUIDField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '{settings.ENTITY_LINK_UID_MAX_LENGTH + 1}' for attribute 'banana' must be between '{settings.ENTITY_LINK_UID_MIN_LENGTH}' and '{settings.ENTITY_LINK_UID_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value='a' * (settings.ENTITY_LINK_UID_MAX_LENGTH + 1)
        )

    def test_min_size(self):
        field = EntityLinkUIDField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '0' for attribute 'banana' must be between '{settings.ENTITY_LINK_UID_MIN_LENGTH}' and '{settings.ENTITY_LINK_UID_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value=''
        )

    def test_empty_is_none(self):
        cls = _EntityLinkUIDFieldTestAuxClass(required='')
        self.assertIsNone(cls.required)

class TestEntityWorkspaceField(TestCase):

    def test_max_size(self):
        field = EntityWorkspaceField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '{settings.ENTITY_WORKSPACE_MAX_LENGTH + 1}' for attribute 'banana' must be between '{settings.ENTITY_WORKSPACE_MIN_LENGTH}' and '{settings.ENTITY_WORKSPACE_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value='a' * (settings.ENTITY_WORKSPACE_MAX_LENGTH + 1)
        )

    def test_min_size(self):
        field = EntityWorkspaceField()
        self.assertRaisesRegex(
            FieldValueError,
            f"The length '0' for attribute 'banana' must be between '{settings.ENTITY_WORKSPACE_MIN_LENGTH}' and '{settings.ENTITY_WORKSPACE_MAX_LENGTH}'.",
            field.validate,
            attr_name='banana',
            value=''
        )

    def test_required_lazy(self):
        cls = _EntityWorkspaceFieldTestAuxClass()
        self.assertRaisesRegex(
            RequiredError,
            "The required attribute is required.",
            cls.validate_required_fields,
        )

    def test_empty_is_none(self):
        cls = _EntityWorkspaceFieldTestAuxClass(required='')
        self.assertIsNone(cls.required)

class TestEntityDateTimeField(TestCase):

    def test_max_size(self):
        value = DateTime(2023, 12, 3, 12, 13, 14)
        with mock.patch('everysk.core.datetime.date_mixin.DateMixin.delta', return_value=DateTime(2023, 12, 2, 12, 13, 14)) as mock_delta:
            field = EntityDateTimeField()
            with self.assertRaises(FieldValueError) as e:
                field.validate(attr_name='banana', value=value)
            self.assertEqual(f"The value '{value}' for field 'banana' must be between {DateTime.market_start()} and {mock_delta.return_value.force_time('LAST_MINUTE')}.", e.exception.msg)

    def test_min_size(self):
        value = DateTime.market_start().delta(-1, 'D')
        with mock.patch('everysk.core.datetime.date_mixin.DateMixin.delta', return_value=DateTime(2023, 12, 2, 12, 13, 14)) as mock_delta:
            field = EntityDateTimeField()
            with self.assertRaises(FieldValueError) as e:
                field.validate(attr_name='banana', value=value)
            self.assertEqual(f"The value '{value}' for field 'banana' must be between {DateTime.market_start()} and {mock_delta.return_value.force_time('LAST_MINUTE')}.", e.exception.msg)

    def test_required_lazy(self):
        cls = _EntityDateTimeFieldTestAuxClass()
        self.assertRaisesRegex(
            RequiredError,
            "The required attribute is required.",
            cls.validate_required_fields,
        )

    def test_empty_is_none(self):
        cls = _EntityDateTimeFieldTestAuxClass(required='')
        self.assertIsNone(cls.required)

    @mock.patch.object(DateTime, 'now', return_value=DateTime(2023, 12, 28, 7, 13, 14))
    def test_validate(self, mock_now: mock.MagicMock):
        entity_cls = _EntityDateTimeFieldTestAuxClass(required='20231229')

        self.assertEqual(entity_cls.required, DateTime(2023, 12, 29, 12, 0, 0))
        with self.assertRaises(FieldValueError) as err:
            entity_cls.required = '20231230'
        self.assertEqual(
            f"The value '{DateTime(2023, 12, 30, 12, 0, 0)}' for field 'required' must be between {DateTime.market_start()} and {DateTime.now().delta(1, 'D').force_time('LAST_MINUTE')}.",
            err.exception.msg
        )
        with self.assertRaises(FieldValueError) as err:
            _EntityDateTimeFieldTestAuxClass(required='20231230')
        self.assertEqual(
            f"The value '{DateTime(2023, 12, 30, 12, 0, 0)}' for field 'required' must be between {DateTime.market_start()} and {DateTime.now().delta(1, 'D').force_time('LAST_MINUTE')}.",
            err.exception.msg
        )

        mock_now.return_value = DateTime(2023, 12, 29, 7, 13, 14)
        entity_cls = _EntityDateTimeFieldTestAuxClass(required='20231230')

        self.assertEqual(entity_cls.required, DateTime(2023, 12, 30, 12, 0, 0))
        with self.assertRaises(FieldValueError) as err:
            entity_cls.required = '20231231'
        self.assertEqual(
            f"The value '{DateTime(2023, 12, 31, 12, 0, 0)}' for field 'required' must be between {DateTime.market_start()} and {DateTime.now().delta(1, 'D').force_time('LAST_MINUTE')}.",
            err.exception.msg
        )
        with self.assertRaises(FieldValueError) as err:
            _EntityDateTimeFieldTestAuxClass(required='20231231')
        self.assertEqual(
            f"The value '{DateTime(2023, 12, 31, 12, 0, 0)}' for field 'required' must be between {DateTime.market_start()} and {DateTime.now().delta(1, 'D').force_time('LAST_MINUTE')}.",
            err.exception.msg
        )

class TestEntityTagsField(TestCase):

    def setUp(self):
        self.field = EntityTagsField()

    def test_clean_value_returns_tagslist_for_none(self):
        cleaned_value = self.field.clean_value(None)
        self.assertIsInstance(cleaned_value, Tags)
        self.assertEqual(len(cleaned_value), 0)

    def test_clean_value_converts_list_to_tagslist(self):
        initial_value = ['tag1', 'tag2']
        cleaned_value = self.field.clean_value(initial_value)
        self.assertIsInstance(cleaned_value, Tags)
        self.assertListEqual(cleaned_value, initial_value)

    def test_clean_value_accepts_tagslist(self):
        tags_list_value = Tags(['tag1', 'tag2'])
        cleaned_value = self.field.clean_value(tags_list_value)
        self.assertIs(cleaned_value, tags_list_value)

    def test_default_value_initialization(self):
        default_tags = ['default_tag']
        field_with_default = EntityTagsField(default=default_tags)
        self.assertIsInstance(field_with_default.default, Tags)
        self.assertListEqual(field_with_default.default, default_tags)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.engines.expression import Expression


###############################################################################
#   Expression Test Case Implementation
###############################################################################
class ExpressionTestCase(TestCase):

    def setUp(self) -> None:
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        return super().setUp()

    ###############################################################################
    #   Get Tokens Method Test Case
    ###############################################################################
    def test_get_tokens(self):
        engine = Expression()
        expected_data = compress({'class_name': 'Expression', 'method_name': 'get_tokens', 'self_obj': engine.to_dict(add_class_path=True), 'params': {'expression': 'a + b', 'data_types': ['cpp_var', 'str_var']}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            engine.get_tokens('a + b')

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    ###############################################################################
    #   Solve Method Test Case
    ###############################################################################
    def test_solve_method(self):
        engine = Expression()
        expected_data = compress({'class_name': 'Expression', 'method_name': 'solve', 'self_obj': engine.to_dict(add_class_path=True), 'params': {'expression': 'a + b', 'user_args': {'a': 1, 'b': 2}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            engine.solve('a + b', {'a': 1, 'b': 2})

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import traceback
from hashlib import sha256
from functools import wraps, _HashedSeq
from time import sleep
from typing import Any, Iterable
from redis import Redis, client, exceptions # pylint: disable=import-self
from redis.backoff import ExponentialBackoff # pylint: disable=import-error, no-name-in-module
from redis.lock import Lock # pylint: disable=import-error, no-name-in-module
from redis.retry import Retry # pylint: disable=import-error, no-name-in-module
from everysk.config import settings
from everysk.core.compress import compress, decompress
from everysk.core.exceptions import RedisEmptyListError
from everysk.core.fields import IntField, StrField
from everysk.core.log import Logger
from everysk.core.object import BaseObject
from everysk.core.serialize import dumps, loads


log = Logger(name='everysk-redis')

## Cache decorator
def _make_key(args: tuple, kwargs: dict) -> str:
    """
    Create a key from args and kwargs to be used on cache.
    This function is a based on functools._make_key

    Args:
        args (tuple): The received args.
        kwargs (dict): The received kwargs.
    """
    #/usr/local/lib/python3.11/functools.py: 448 - def _make_key
    key = args
    if kwargs:
        for item in kwargs.items():
            key += item
    elif len(key) == 1 and type(key[0]) in {int, float, str}:
        # To be faster if we have only one item and it's a int, float or str
        # we return it directly
        return str(key[0])

    return str(_HashedSeq(key))

def cache(*, timeout: int):
    """
    Decorator to cache the result of a function in Redis.
    To disable the cache, set the timeout to None otherwise it must be > 0.

    Args:
        timeout (int): The time in seconds that the result will be stored in cache.

    Raises:
        ValueError: If timeout is not an integer or is less than 0.
    """
    if timeout is not None and (not isinstance(timeout, int) or timeout < 1):
        raise ValueError('Timeout must be an integer greater than 0.')

    # Because we have a parameter we need to create another nested function
    def decorator(func):
        info = {'hits': 0, 'misses': 0}
        redis_cache = RedisCache(prefix=func.__name__)

        # We use wraps to keep the original function name and docstring
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = _make_key(args, kwargs)
            value = redis_cache.get(key)
            if value:
                info['hits'] += 1
                return loads(value, protocol='pickle')

            info['misses'] += 1
            value = func(*args, **kwargs)
            redis_cache.set(key, dumps(value, protocol='pickle'), timeout)
            return value

        # We add some extra attributes to the wrapper
        wrapper.info = info
        wrapper.clear = redis_cache.delete_prefix

        return wrapper

    return decorator


###############################################################################
#   RedisClient Class Implementation
###############################################################################
class RedisClient(BaseObject):
    ## Private attributes
    _connection: Redis = None
    _separator = ':'

    ## Public attributes
    prefix: str = None
    host: str = None
    port: int = None

    ## Private methods
    def _connect(self) -> None:
        """ Create a Redis connection and stores to later use. """
        # https://redis-py.readthedocs.io/en/stable/retry.html
        retry = Retry(ExponentialBackoff(), 3)

        # https://github.com/redis/redis-py/issues/722
        # We use RedisClient._connection to create a Singleton connection
        log.debug('Connecting on Redis(%s).....', self.host)
        RedisClient._connection = Redis(
            host=self.host,
            port=self.port,
            health_check_interval=30, # seconds
            socket_keepalive=True,
            socket_timeout=120, # seconds
            retry=retry,
            retry_on_error=[Exception]
        )

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.host is None:
            self.host = settings.REDIS_HOST

        if self.port is None:
            self.port = settings.REDIS_PORT

        if RedisClient._connection is None:
            self._connect()

    @property
    def connection(self) -> Redis:
        """
        We use this property to check if Redis is online
        then returning the working connection.
        """
        try:
            RedisClient._connection.ping()
        except Exception: # pylint: disable=broad-exception-caught
            # Create a new connection
            self._connect()

        return RedisClient._connection

    def flush_all(self) -> bool:
        """ Clear all keys from Redis. """
        return self.connection.flushall()

    def get_hash_key(self, key: bytes | str) -> str:
        """
        Convert the key to Sha256 to avoid strange characters and decrease the chance of generating the same key.

        Args:
            key (bytes | str): The key to be hashed.
        """
        # To be used by SHA256 the key must be bytes
        if isinstance(key, str):
            key = key.encode('utf-8')

        key = sha256(key).hexdigest()
        if self.prefix:
            key = f'{self.prefix}{self._separator}{key}'

        return key


###############################################################################
#   RedisCache Class Implementation
###############################################################################
class RedisCache(RedisClient):
    """ Redis cache client """

    def get_set(self, key: str, func: callable, timeout: int = None, **kwargs) -> Any:
        """
        This method execute a get on Redis server, if this returns None then we execute the set method.
        For the set method we use the func(**kwargs).
        To avoid race conditions we use a RedisLock to run only one set method.

        Args:
            key (str): The key that will be used to cache the result from func.
            func (callable): The function that generates the desired cached result.
            timeout (int, optional): The timeout that this key will be keep on cache. Defaults to None.
            *kwargs (dict, optional): Extra params that will be send to the func.

        Returns:
            Any: The result from cache get or the result from func.
        """
        result = self.get(key)
        if result is None:
            # We create a lock on Redis
            lock = RedisLock(name=f'redis-get-set-lock-{key}')
            if lock.acquire():
                # If we can acquire the lock then we don't have race conditions
                # and proceed normally
                try:
                    result = func(**kwargs)
                    self.set(key=key, value=result, timeout=timeout)
                except Exception: # pylint: disable=broad-exception-caught
                    # We generate a log for this exception
                    log.error('Redis get_set method: %s', traceback.format_exc())

                # Then we release the the lock
                lock.release()

            else:
                # If we can't acquire the lock that means we have race condition
                # in this case we need to wait for the key be set or if some error
                # occur the lock will be released.
                result = self.get(key)
                while result is None:
                    result = self.get(key)
                    if result is None and lock.acquire():
                        # Then we release the the lock and exit
                        lock.release()
                        break

                    # We wait 0.5 second until next try
                    sleep(0.5)

        if isinstance(result, bytes):
            result = result.decode('utf-8')

        return result

    def get(self, key: bytes) -> bytes:
        """
        Get the value from key from connection, if it's not found return None.

        Returns:
            bytes: Always return a byte object.
        """
        key = self.get_hash_key(key)
        value = self.connection.get(key)
        return value

    def set(self, key: bytes, value: Any, timeout: int = None) -> None:
        """
        Set key/value on connection for timeout in seconds,
        if timeout is None the key/value will be keep forever.
        Value must be one of these: bytes, str, int or float.
        """
        if not isinstance(value, (bytes, str, int, float)):
            raise ValueError(
                f'Value {type(value)} is not in supported values. (bytes, str, int, float).'
            )

        key = self.get_hash_key(key)
        self.connection.set(key, value, timeout)

    def delete(self, key: bytes | str | Iterable) -> None:
        """
        Delete one or more keys from Redis.

        Args:
            key (bytes | str | Iterable): The key or keys to be deleted.
        """
        if not isinstance(key, (list, set, tuple)):
            keys = [self.get_hash_key(key)]
        else:
            keys = [self.get_hash_key(k) for k in key]

        self.connection.delete(*keys)

    def delete_prefix(self, prefix: str = None) -> None:
        """
        Delete all keys from Redis that start with the given prefix or with the one set in the class.

        Args:
            prefix (str, optional): The same string that was used to create the keys. Defaults to None.
        """
        # If prefix is None, we use the prefix from the class
        prefix = prefix or self.prefix
        # To search for all keys with this prefix we need to add a ':*' at the end
        if '*' not in prefix:
            prefix = f'{prefix}{self._separator}*'

        # We use the scan method to search for all keys with this prefix and pipe to delete them
        pipe = self.connection.pipeline()
        cursor = None
        while cursor != 0:
            cursor, keys = self.connection.scan(cursor=cursor or 0, match=prefix, count=1000)
            if keys:
                pipe.delete(*keys)

        pipe.execute()


###############################################################################
#   RedisCacheCompressed Class Implementation
###############################################################################
class RedisCacheCompressed(RedisCache):
    """
    Store data on Redis server using pickle and zlib
    Use this if you need to store objects ons Redis.
    """

    def get(self, key: bytes) -> Any:
        """
        Retrieve the value associated with the given key from the Redis cache

        Args:
            key (bytes): The key to retrieve the value for

        """
        value = super().get(key)
        if value is not None:
            value = decompress(value, serialize='pickle')

        return value

    def set(self, key: bytes, value: Any, timeout: int = None) -> None:
        """
        Store the provided value in the Redis cache associated with the given key

        Args:
            key (bytes): The key to associate with the value
            value (Any): The value to store in the Redis cache
            timeout (int, optional): Expiration time for the cached item
        """
        value = compress(value, serialize='pickle')
        super().set(key, value, timeout)

###############################################################################
#   RedisList Class Implementation
###############################################################################
class RedisList(RedisClient):
    """
    First in, first out Redis list implementation.
    -> https://redis.io/docs/data-types/lists/
    -> https://koalatea.io/python-redis-lists/

    """
    name = StrField(required=True)

    def bpop(self, timeout: int = 0) -> tuple:
        """
        Pop the first item from the list, blocking until a item exists
        or timeout was reached.
        If timeout is 0, then block indefinitely.

        Returns:
            tuple: (list name, value)
        """
        value = self.connection.blpop(self.name, timeout=timeout)
        if value is None:
            raise RedisEmptyListError(f"The RedisList(name='{self.name}') is empty.")

        name, value = value
        if isinstance(name, bytes):
            name = name.decode()

        return (name, decompress(value, serialize='pickle'))

    def pop(self) -> Any:
        """
        Pop the first item from the list.

        Raises:
            RedisEmptyListError: If the return is None/empty.
        """
        value = self.connection.lpop(self.name)
        if value is None:
            raise RedisEmptyListError(f"The RedisList(name='{self.name}') is empty.")

        value = decompress(value, serialize='pickle')
        return value

    def push(self, value: Any) -> None:
        """
        Puts value on the last position of the list.

        Args:
            value (Any): the value to be inserted into the last position
        """
        value = compress(value, serialize='pickle')
        self.connection.rpush(self.name, value)

    def clear(self) -> None:
        """
        Clear all keys
        """
        self.connection.delete(self.name)

###############################################################################
#   RedisChannel Class Implementation
###############################################################################
class RedisChannel(RedisClient):
    """
    Base class to work with channels on Redis.
    https://blog.devgenius.io/how-to-use-redis-pub-sub-in-your-python-application-b6d5e11fc8de
    """
    _channel: client.PubSub = None
    exit_message = StrField(default='exit', readonly=True)
    name = StrField(required=True)

    def send(self, message: dict) -> None:
        self.connection.publish(self.name, message)

    @property
    def channel(self) -> client.PubSub:
        """ Create a connection with name """
        if self._channel is None:
            self._channel = self.connection.pubsub()
            self._channel.subscribe(self.name)

        return self._channel

    def parse_message(self, message: dict) -> tuple:
        """
        Convert message data from bytes to str

        Args:
            message (dict): The message dictionary to be parsed.

        Returns:
            tuple: Containing the channel name and its corresponding data.
        """
        # message format
        # {'type': None, 'pattern': None, 'channel': None, 'data': None}
        channel_name = None
        data = None
        if message:
            channel_name = message.get('channel', None) or None
            data = message.get('data', '') or ''
            if isinstance(channel_name, bytes):
                channel_name = channel_name.decode()
            if isinstance(data, bytes):
                data = data.decode()

        return (channel_name, data)

    def consume(self, callback: callable = None) -> None:
        """ Loop for consume message from channel when they arrive. """
        for message in self.channel.listen():
            channel_name, data = self.parse_message(message)
            # Only care if the message is sent to this channel
            if channel_name == self.name:
                # Stop iteration on exit_message
                if data == self.exit_message:
                    break

                # We can use a function for callback or self.process_message
                if callback:
                    callback(data)
                else:
                    self.process_message(data)

    def process_message(self, message: str) -> None:
        """ Use it on child classes to manipulate the received message. """
        pass

###############################################################################
#   RedisLock Class Implementation
###############################################################################
class RedisLock(RedisClient):
    """
    Class used to create a lock on Redis
    https://rohansaraf.medium.com/distributed-locking-with-redis-ecb0773e7695
    https://redis-py.readthedocs.io/en/latest/lock.html
    """
    ## Private attributes
    _lock = None

    ## Public attributes
    name = StrField(required=True)
    timeout = IntField(default=None) # timeout indicates a maximum life for the lock in seconds.

    ## Private methods
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self._lock is None:
            self._lock = Lock(redis=self.connection, name=self._get_name(), timeout=self.timeout)

    def _get_name(self) -> str:
        """
        Convert self.name to a SHA256 hash, this avoid strange chars on name that can broke Redis.
        """
        return self.get_hash_key(self.name)

    ## Public methods
    def acquire(self, blocking: bool = False, blocking_timeout: float = None) -> bool:
        """
        Try to acquire a lock with self.name, if lock is already acquired returns False.
        If blocking is False, always return immediately, if blocking is True it will waiting until block can be acquired.
        blocking_timeout specifies the maximum number of seconds to wait trying to acquire the lock.
        """
        return self._lock.acquire(blocking=blocking, blocking_timeout=blocking_timeout)

    def owned(self) -> bool:
        """ Returns True if this key is locked by this lock, otherwise False. """
        return self._lock.owned()

    def release(self, force: bool = False) -> bool:
        """
        Try to release this lock, will be True only if this lock is owned by this process.
        If force is True, the lock will be released even if it is not owned.
        """
        if force:
            self.connection.delete(self._get_name())
            return True

        try:
            self._lock.release()
            return True
        except (exceptions.LockError, exceptions.LockNotOwnedError):
            pass

        return False



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['FileResponse', 'HTMLResponse', 'JSONResponse', 'PlainTextResponse', 'RedirectResponse', 'Response', 'StreamingResponse']
from typing import Any
from starlette.responses import Response, FileResponse, HTMLResponse, PlainTextResponse, RedirectResponse, StreamingResponse

from everysk.core.serialize import dumps


class JSONResponse(Response):
    media_type = 'application/json; charset=UTF-8'

    def render(self, content: Any) -> bytes:
        """
        Serialize the content to JSON format using the dumps function from the core.serialize module.

        Args:
            content (Any): The content to be serialized.
        """
        # with indent=None and separators=(",", ":"), the JSON will be minified
        return dumps(content, indent=None, separators=(",", ":"), protocol='json', use_undefined=True).encode('utf-8')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import IntField, StrField, RegexField, SetField

###############################################################################
#   Settings Implementation
###############################################################################
ENTITY_WORKSPACE_MAX_LENGTH = IntField(default=50, readonly=True)
ENTITY_WORKSPACE_MIN_LENGTH = IntField(default=1, readonly=True)
ENTITY_WORKSPACE_REGEX = RegexField(default=r'^[a-zA-Z0-9_]*$', readonly=True)

ENTITY_NAME_MAX_LENGTH = IntField(default=200, readonly=True)
ENTITY_NAME_MIN_LENGTH = IntField(default=1, readonly=True)

ENTITY_DESCRIPTION_MAX_LEN = IntField(default=500, readonly=True)
ENTITY_DESCRIPTION_MIN_LEN = IntField(default=0, readonly=True)

ENTITY_MAX_TAG_LENGTH = IntField(default=252, readonly=True)
ENTITY_MIN_TAG_LENGTH = IntField(default=1, readonly=True)
ENTITY_MAX_TAG_SIZE = IntField(default=32, readonly=True)
ENTITY_MIN_TAG_SIZE = IntField(default=0, readonly=True)

ENTITY_DEFAULT_VERSION = StrField(default='v1', readonly=True)
ENTITY_ID_LENGTH = IntField(default=25, readonly=True)

ENTITY_MAX_TAG_SIZE = IntField(default=32, readonly=True)
ENTITY_MIN_TAG_SIZE = IntField(default=0, readonly=True)

ENTITY_LINK_UID_MAX_LENGTH = IntField(default=128, readonly=True)
ENTITY_LINK_UID_MIN_LENGTH = IntField(default=1, readonly=True)

DEFAULT_QUERY_OFFSET = IntField(default=0, readonly=True)
DEFAULT_QUERY_PAGE_SIZE = IntField(default=20, readonly=True)
DEFAULT_QUERY_LIMIT = IntField(default=None)

QUERY_OPERATORS = SetField(default={'<=', '>=', '<', '>', '=', '!=', 'IN', 'NOT_IN'}, readonly=True)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['Middleware', 'GZipMiddleware', 'SecurityHeadersMiddleware']
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.gzip import GZipMiddleware

from everysk.config import settings


GZIP_MINIMUM_SIZE = settings.EVERYSK_SERVER_GZIP_MINIMUM_SIZE
GZIP_COMPRESS_LEVEL = settings.EVERYSK_SERVER_GZIP_COMPRESS_LEVEL


def update_with_default_middlewares(middlewares: list[Middleware]) -> list[Middleware]:
    """
    Update the given middleware list with the default ones.
    The default middleware are the GZipMiddleware and SecurityHeadersMiddleware.
    """
    if middlewares is None:
        middlewares = []

    middlewares.insert(0, Middleware(GZipMiddleware, minimum_size=GZIP_MINIMUM_SIZE, compresslevel=GZIP_COMPRESS_LEVEL))
    middlewares.append(Middleware(SecurityHeadersMiddleware))
    return middlewares


class SecurityHeadersMiddleware(BaseHTTPMiddleware):
    """
    Middleware to add security headers to the response.
    These headers are used to protect the application against
    some types of attacks and will be added to every response.
    """

    async def dispatch(self, request, call_next):
        response = await call_next(request)

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security
        response.headers['Strict-Transport-Security'] = 'max-age=31536000; includeSubDomains'

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Content-Type-Options
        response.headers['X-Content-Type-Options'] = 'nosniff'

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-DNS-Prefetch-Control
        response.headers['X-DNS-Prefetch-Control'] = 'off'

        # https://webtechsurvey.com/response-header/x-download-options
        # Only works on IE8
        response.headers['X-Download-Options'] = 'noopen'

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options
        response.headers['X-Frame-Options'] = 'DENY'

        # https://webtechsurvey.com/response-header/x-permitted-cross-domain-policies
        response.headers['X-Permitted-Cross-Domain-Policies'] = 'none'

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-XSS-Protection
        # Warning: Even though this feature can protect users of older web browsers that don't yet support CSP, in some cases,
        # XSS protection can create XSS vulnerabilities in otherwise safe websites. See the section below for more information.
        # response.headers['X-XSS-Protection'] = '1; mode=block'

        return response



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from contextlib import AbstractContextManager
from contextvars import ContextVar
from copy import deepcopy
from importlib import import_module
from inspect import getmembers, isroutine, isclass
from os import getenv
from os.path import dirname, exists, sep
from pathlib import Path, PosixPath
from re import findall
from types import TracebackType, UnionType
from typing import Any
from everysk.core.object import BaseObject, MetaClass
from everysk.utils import bool_convert

# The root path of the everysk lib we need this to create the correct stub file
EVERYSK_ROOT = dirname(__file__)

## Private functions
def _get_context_value(name: str) -> Any:
    """
    If we are in a python context we return the value
    that is in the context setting.
    If the settings is not found we raise a KeyError.

    Args:
        name (str): The settings name.

    Raises:
        KeyError: If the name is not found.
    """
    context_settings = SettingsManager.settings.get()
    # We only care for the values that are explicit set in this settings
    if context_settings:
        return context_settings.__dict__.get(name, Undefined)

    return Undefined


def _get_env_value__(name: str) -> Any:
    """
    Get the value from the environment and clean it.

    Args:
        name (str): The var name to search.
    """
    env_value = getenv(name)

    # This is a check to always get the correct value
    # that we received from the environment
    # https://everysk.atlassian.net/browse/COD-907
    if env_value is not None:
        attributes = getattr(Settings, SettingsMetaClass._attr_name, {}) # pylint: disable=protected-access
        try:
            attr_type = attributes[name]
            try:
                # Our fields have the clean_value method
                env_value = attr_type.clean_value(env_value)

            except AttributeError:
                # For the other types we try to convert the
                # value to the specific type that was declared
                # For bool types we use a specific function
                if attr_type != bool:
                    env_value = attr_type(env_value)
                else:
                    env_value = bool_convert(env_value)

        except KeyError:
            pass
    else:
        # If env_value is None then the var was not set
        env_value = Undefined

    return env_value

def _is_valid_path(path: str) -> bool:
    """
    Function that checks if a path is valid to load the settings module.

    Args:
        path (str): The full path that need be checked.
    """
    result = True
    # Python installation normally happens on venvs that are in the project path
    if 'site-packages/' in path:
        result = False

    # Git files could have copies of python files
    elif '.git/' in path:
        result = False

    return result


## Public functions
def get_full_dotted_path_module(entry: PosixPath, root: str) -> str:
    """
    From the full path we need to generate the module that is valid for import.
    Changes '/var/app/src/everysk/settings.py' to 'everysk.settings'.

    Args:
        entry (PosixPath): The posix path object.
        root (str): The root path that will be removed.
    """
    # /var/app/src/everysk/settings.py
    module = str(entry)

    # everysk/settings.py
    module = module.replace(f'{root}{sep}', '')

    # everysk/settings
    module = module.replace('.py', '')

    # everysk.settings
    module = module.replace(sep, '.')

    return module

def get_all_modules() -> list:
    """
    We search inside EVERYSK_ROOT and PROJECT_ROOT for 'settings.py'
    files and convert this files into python modules.

    Returns:
        list: A list of Python modules

    Example:
        >>> all_modules = get_all_modules()
        >>> print(all_modules)
        ... ['module1', 'module2']
    """
    search_name = 'settings.py'
    modules = []
    if EVERYSK_ROOT and exists(EVERYSK_ROOT):
        # we need to remove the '/everysk' from the path to create the correct module path
        root = EVERYSK_ROOT.replace(f'{sep}everysk', '')
        for entry in Path(EVERYSK_ROOT).rglob(search_name):
            modules.append(get_full_dotted_path_module(entry=entry, root=root))

    project_root = getenv('PROJECT_ROOT')
    if project_root and exists(project_root):
        for entry in Path(project_root).rglob(search_name):
            # The rglob function does not have a way to remove some paths
            # so we need to manually exclude the ones that are not required
            if _is_valid_path(path=entry.as_posix()):
                modules.append(get_full_dotted_path_module(entry=entry, root=project_root))

    return modules

def get_all_attributes() -> list[tuple[str, Any, Any]]:
    """
    Get all attributes from the settings modules except ones that starts with '_'

    Example:
        Consider a settings module 'my_settings.py' containing the following attributes:

        ...
        MY_SETTING_1 = 42
        MY_SETTING_2 = 'Hello, World!'
        _MY_SETTING_3 = 'Foo Bar'
        ...

        Calling get_all_attributes() will return:

        [
            ('MY_SETTING_1', int, 42),
            ('MY_SETTING_2', str, 'Hello, World!')
        ]
    """
    result = []

    for module in get_all_modules():
        # Import the module
        module = import_module(module)

        for attr, value in getmembers(module, predicate=lambda x: not isroutine(x) and not isclass(x)):
            if not attr.startswith('_'):
                try:
                    attr_value = value.default
                    attr_type = value
                except AttributeError:
                    attr_value = value
                    # If this attribute already has an annotation on the module we use it
                    # otherwise we get the class of the value as the annotation
                    # https://everysk.atlassian.net/browse/COD-3833
                    attr_type = module.__annotations__.get(attr, type(value))

                result.append((attr, attr_type, attr_value))
    return result


def update_settings_attributes(obj: 'Settings'):
    """
    Updates the attributes of the Settings class in the config.pyi file.

    This function updates the attribute definitions of the Settings class based on the attributes retrieved from the `get_all_attributes` function.

    Example:
        >>> from everysk.config import update_settings_attributes
        >>> update_settings_attribute()
        ...
    """
    with open(f'{EVERYSK_ROOT}/config.pyi', 'w', encoding='utf-8') as stubs:
        # This fix the correct type of SettingsManager.settings
        stubs.write('from contextvars import ContextVar\n')

        # This fix the stub file for the RegexField
        stubs.write('from re import Pattern\n\n')

        # Now we have dict and BaseDict in the stub file too
        stubs.write('from everysk.core.object import BaseDict\n\n')

        # Write the SettingsManager class
        stubs.write('class SettingsManager:\n')
        stubs.write('    settings: ContextVar\n')
        stubs.write('    token: str\n\n')

        # Write the Settings class
        stubs.write('class Settings:\n')

        attributes = set()
        for attr_name, attr_type, attr_value in get_all_attributes():
            obj.__set_attribute__(attr_name, attr_type, attr_value)
            try:
                # If attr_type is a Everysk Field the real value will be in the attr_type attribute
                attr_type = attr_type.attr_type
            except AttributeError:
                pass

            # For Union types we need to get the real types inside __args__
            if not isinstance(attr_type, UnionType):
                attr_type = attr_type.__name__
            else:
                attr_type = ' | '.join([t.__name__ for t in attr_type.__args__])

            # We could have some duplicated attributes so we keep then in a set
            attributes.add(f'    {attr_name}: {attr_type}\n')

        # Then we write all attributes in the stub file
        stubs.writelines(sorted(attributes))
        # To autocomplete the init params
        stubs.write('    def __init__(self, singleton: bool = True, **kwargs) -> None: ...\n')
        # Fix for the autocomplete works if a context var is created
        stubs.write('    def __enter__(self) -> \'Settings\': ...\n')

        # This is needed to VS Code understand the instance
        stubs.write('\nsettings: Settings\n')


## Public classes
class SettingsManager(AbstractContextManager):
    settings: ContextVar = ContextVar('everysk-settings', default=None)
    token: str = None

    def __init__(self, context_settings: 'Settings' = Undefined) -> None:
        if context_settings:
            # We make a copy to not change the original
            context_settings = deepcopy(context_settings)
        else:
            # We create a fresh one
            context_settings = Settings(singleton=False)

        # Store the value inside the context var
        self.token = self.settings.set(context_settings)

    def __exit__(self, __exc_type: type[BaseException] | None, __exc_value: BaseException | None, __traceback: TracebackType | None) -> bool | None:
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__exit__

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        # Clear all settings config exiting the context
        self.settings.reset(self.token)

        # return False to raise any errors if they happened
        return False


class SettingsMetaClass(MetaClass):

    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> type:
        """
        Method that create the class object.

        Args:
            mcs (_type_): This class.
            name (str): The name of the new class -> Settings.
            bases (tuple): Parent for the new class -> BaseObject.
            attrs (dict): A list of attributes tha the new class will have.
        """
        obj = super().__new__(mcs, name, bases, attrs)

        # This is executed only one time for every python process
        # before the initialization of the instance, we update the attributes list
        update_settings_attributes(obj)

        return obj


class Settings(BaseObject, metaclass=SettingsMetaClass):

    def __new__(cls, singleton: bool = True, **kwargs) -> 'Settings':
        """
        Changed to keep only one instance for the class.

        Args:
            singleton (bool): Used to keep the same instance. Defaults to True.
        """
        if singleton:
            try:
                return settings
            except NameError:
                pass

        return object.__new__(cls)

    def __init__(self, **kwargs) -> None:
        # remove the singleton keyword
        if 'singleton' in kwargs:
            kwargs.pop('singleton')

        super().__init__(**kwargs)

    def __deepcopy__(self, memo: dict = None) -> 'Settings':
        """
        A deep copy constructs a new compound object and then, recursively,
        inserts copies into it of the objects found in the original.
        This method is used when we call deepcopy(obj).

        Args:
            memo (dict, optional): A memory object to avoid copy twice. Defaults to None.
        """
        # We need to copy the __dict__
        obj = deepcopy(self.__dict__, memo)

        # We create a new obj
        obj = type(self)(singleton=False, **obj)
        return obj

    def __enter__(self) -> 'Settings':
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__enter__

        Returns:
            Settings: A copy of the settings object.
        """
        # We must use deepcopy to keep all values that are already set
        return deepcopy(settings)

    def __exit__(self, __exc_type: type[BaseException] | None, __exc_value: BaseException | None, __traceback: TracebackType | None) -> bool | None:
        """
        https://docs.python.org/3/library/stdtypes.html#contextmanager.__exit__
        This method is required even if it only returns False

        Returns:
            bool | None: If return is False any exception will be raised.
        """
        return False

    def __getattribute__(self, name: str) -> Any:
        """
        This method try to first get the value from the environment
        if this does not exists get from the class.

        Args:
            name (str): The setting name.
        """
        ## For performance and to avoid some errors because
        ## we are trying to get other things that does not
        ## are attributes of this class
        if name.startswith('__') or name == '_config':
            return super().__getattribute__(name)

        # First we try to get the value from the context
        value = _get_context_value(name)
        if value is Undefined:
            # Second we try to get the value from the instance
            value = self.__dict__.get(name, Undefined)
            if value is Undefined:
                # Third we try to get from the environment
                value = _get_env_value__(name)
                if value is Undefined:
                    # Fourth we try to get from the class that is the default value
                    value = super().__getattribute__(name)

        if isinstance(value, str):
            try:
                # If is a normal string nothing happen
                value = value.format()
            except KeyError:
                # Otherwise is a string in this format:
                # '{SOME_SETTING}-rest-string' so we need to change SOME_SETTING
                # for the real value that is a settings

                # We get all {*} that appears
                keys = findall(r'{(.*?)}', value)
                # Then we get the real value of each key
                kwargs = {key: getattr(self, key) for key in keys}
                # Then we try to format the real string
                value = value.format(**kwargs)

        return value

    def __setattr__(self, name: str, value: Any) -> None:
        """
        This method set the value of every attribute inside the Settings object,
        we use it to choose if we set on the normal object or in the context object.
        If the context_settings exists it sets the value inside the context_settings.

        Args:
            name (str): The name of the attribute.
            value (Any): The value that needs to be stored.
        """
        context_settings = SettingsManager.settings.get()
        # If context settings does not exist or it is a context_settings
        # that is trying to store the value for this setting we use the default behavior
        if not context_settings or context_settings == self:
            super().__setattr__(name, value)
        else:
            # Otherwise we call the set for context_settings
            setattr(context_settings, name, value)
            SettingsManager.settings.set(context_settings)


## Here we load all computed settings on one var
settings: Settings = Settings()



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
import json
import logging
import traceback
import sys
from logging import Logger as PythonLogger, DEBUG, StreamHandler
from everysk.config import settings
from everysk.core import log as log_module
from everysk.core.unittests import TestCase, skipIf, mock
try:
    from flask import Flask
except ImportError:
    Flask = None


class LoggerJsonTestCase(TestCase):
    # The first TestCase so we don't need to change the line number in the tests
    # every time we add a new test or lines

    def setUp(self) -> None:
        self.old_logging_json = settings.LOGGING_JSON
        settings.LOGGING_JSON = True
        self.log = log_module.Logger(name='everysk-log-json-test')

    def tearDown(self) -> None:
        settings.LOGGING_JSON = self.old_logging_json

    def test_info_as_json(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('info', stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "info", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 39, "function": "test_info_as_json"}}\n'
            )

    def test_info_as_json_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('info with labels', extra={'labels': {'key': 'value'}}, stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "info with labels", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {"key": "value"}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 56, "function": "test_info_as_json_with_labels"}}\n'
            )

    def test_info_as_json_with_labels_as_bytes(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('info with labels as bytes', extra={'labels': b'text'}, stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "info with labels as bytes", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": "text", '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 73, "function": "test_info_as_json_with_labels_as_bytes"}}\n'
            )

    def test_info_as_json_with_labels_with_bytes(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('info with labels with bytes', extra={'labels': {'key': b'text'}}, stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "info with labels with bytes", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {"key": "text"}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 90, "function": "test_info_as_json_with_labels_with_bytes"}}\n'
            )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_info_as_json_with_http_headers(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('info with http headers', extra={'http_headers': {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}}, stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "info with http headers", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                '"traceback": "", '
                '"http": {"headers": {"traceparent": "00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01"}, "payload": {}}, '
                '"logging.googleapis.com/trace": "projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df", '
                '"logging.googleapis.com/spanId": "1c6c592f9e46e3fb", '
                '"logging.googleapis.com/trace_sampled": true, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 108, "function": "test_info_as_json_with_http_headers"}}\n'
            )

    def test_error_as_json(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            trace = ''
            try:
                raise ValueError('test')
            except ValueError:
                trace = traceback.format_exc()
                self.log.error('error', stacklevel=3)
            stream.write.assert_called_once_with(
                '{"message": "error", '
                '"severity": "ERROR", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                f'"traceback": {json.dumps(trace)}, '
                '"http": {"headers": {}, "payload": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 130, "function": "test_error_as_json"}}\n'
            )


@skipIf(Flask is None, 'Flask is not installed')
class LoggerFlaskTestCase(TestCase):

    def setUp(self) -> None:
        self.app = Flask(__name__)

    def test_flask_headers(self):
        from werkzeug.datastructures import Headers # pylint: disable=import-outside-toplevel
        self.assertDictEqual(log_module._get_flask_headers(), {})
        with self.app.test_request_context('/') as ctx:
            ctx.request.headers = Headers([('X-Request-ID', '123')])
            self.assertEqual(log_module._get_flask_headers(), {'x-request-id': '123'})

    def test_flask_headers_runtime_error(self):
        with self.app.test_request_context('/') as ctx:
            ctx.request.headers = mock.MagicMock()
            ctx.request.headers.items.side_effect = RuntimeError
            self.assertDictEqual(log_module._get_flask_headers(), {})

    def test_flask_payload(self):
        self.assertDictEqual(log_module._get_flask_payload(), {})
        with self.app.test_request_context('/') as ctx:
            ctx.request._cached_data = b'{"key": "value"}'
            self.assertEqual(log_module._get_flask_payload(), {'key': 'value'})

    def test_flask_payload_runtime_error(self):
        with self.app.test_request_context('/') as ctx:
            ctx.request.headers = mock.MagicMock()
            ctx.request.headers.items.side_effect = RuntimeError
            self.assertDictEqual(log_module._get_flask_payload(), {})

    def test_flask_payload_invalid_json(self):
        with self.app.test_request_context('/') as ctx:
            ctx.request._cached_data = b'invalid json'
            self.assertDictEqual(
                log_module._get_flask_payload(),
                {'body': b'invalid json', 'message': 'The payload could not be decoded as JSON.'}
            )


class LoggerTraceTestCase(TestCase):

    def setUp(self) -> None:
        self.trace_parts = log_module.DEFAULT_TRACE_PARTS.copy()

    def test_traceparent(self):
        traceparent = '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'
        log_module._parse_traceparent(traceparent, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': '4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    def test_parse_traceparent_invalid(self):
        traceparent = 'invalid'
        log_module._parse_traceparent(traceparent, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '',
                'trace_id': '',
                'trace_sample': False,
                'version': '',
            }
        )

    def test_parse_x_cloud_trace_context(self):
        trace_context = '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'
        log_module._parse_x_cloud_trace_context(trace_context, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '2048109991600514043',
                'trace_id': '4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '',
            }
        )

    def test_parse_x_cloud_trace_context_invalid(self):
        trace_context = 'invalid'
        log_module._parse_x_cloud_trace_context(trace_context, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {'span_id': '', 'trace_id': '', 'trace_sample': False, 'version': ''}
        )

    def test_get_trace_data_no_headers(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={}),
            {'span_id': '', 'trace_id': '', 'trace_sample': False, 'version': ''}
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_traceparent(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}),
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_x_cloud_trace_context(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'}),
            {
                'span_id': '2048109991600514043',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '',
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_both(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={
                'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01',
                'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'
            }),
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    def test_get_traceback(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertEqual(log_module._get_traceback(), traceback.format_exc())

    def test_get_traceback_no_traceback(self):
        self.assertEqual(log_module._get_traceback(), '')


class LoggerManagerTestCase(TestCase):

    def test_http_headers(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'http_headers': {'X-Request-ID': '123'}})

    def test_http_payload(self):
        with log_module.LoggerManager(http_payload={'data': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'http_payload': {'data': '123'}})

    def test_labels(self):
        with log_module.LoggerManager(labels={'data': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'labels': {'data': '123'}})

    def test_stacklevel(self):
        with log_module.LoggerManager(stacklevel=1):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'stacklevel': 1})

    def test_traceback(self):
        trace = ''
        try:
            raise ValueError('test')
        except ValueError:
            trace = traceback.format_exc()

        with log_module.LoggerManager(traceback=trace):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'traceback': trace})

    def test_empty(self):
        with log_module.LoggerManager():
            self.assertEqual(log_module.LoggerManager._extra.get(), {})

    def test_chain_managers(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            with log_module.LoggerManager(http_payload={'data': '123'}):
                with log_module.LoggerManager(labels={'data': '123'}):

                    self.assertEqual(
                        log_module.LoggerManager._extra.get(),
                        {'http_headers': {'X-Request-ID': '123'}, 'http_payload': {'data': '123'}, 'labels': {'data': '123'}}
                    )
                self.assertEqual(
                    log_module.LoggerManager._extra.get(),
                    {'http_headers': {'X-Request-ID': '123'}, 'http_payload': {'data': '123'}}
                )
            self.assertEqual(
                log_module.LoggerManager._extra.get(),
                {'http_headers': {'X-Request-ID': '123'}}
            )

        self.assertEqual(log_module.LoggerManager._extra.get(), {})

    def test_reset(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            log_module.LoggerManager.reset()
            self.assertEqual(log_module.LoggerManager._extra.get(), {})


class LoggerTestCase(TestCase):

    def test_logger_return(self):
        log = log_module.Logger('everysk-log-stdout-test')
        self.assertIsInstance(log._log, PythonLogger)

    def test_log_level(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        self.assertEqual(log._log.level, DEBUG)

    def test_log_propagate(self):
        log = log_module.Logger('everysk-log-stdout-test')
        self.assertFalse(log._log.propagate)

    def test_stdout_handler(self):
        log = log_module.Logger('everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(type(handler), StreamHandler)

    def test_handler_level(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(handler.level, DEBUG)

    @mock.patch.dict('os.environ', {'LOGGING_JSON': 'false'})
    def test_handler_format(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(
            handler.formatter._fmt,
            '%(asctime)s - %(levelname)s - %(labels)s - %(message)s'
        )

    def test_default_root_name(self):
        # https://everysk.atlassian.net/browse/COD-3212
        with self.assertRaisesRegex(ValueError, 'The name of the log could not be "root".'):
            log_module.Logger(name='root')


class LoggerExtraDataTestCase(TestCase):

    def setUp(self) -> None:
        self.log = log_module.Logger('everysk-log-extra-test')

    def test_http_headers(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'X-Request-ID': '123'}}),
            {'labels': {}}
        )

    def test_http_headers_traceparent(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'traceparent': '123'}}),
            {'http_headers': {'traceparent': '123'}, 'labels': {}}
        )

    def test_http_headers_x_cloud_trace_context(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'x-cloud-trace-context': '123'}}),
            {'http_headers': {'x-cloud-trace-context': '123'}, 'labels': {}}
        )

    def test_http_payload(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_payload': {'data': '123'}}),
            {'http_payload': {'data': '123'}, 'labels': {}}
        )

    def test_another_key(self):
        self.assertDictEqual(
            self.log._get_extra_data({'another_key': '123'}),
            {'another_key': '123', 'labels': {}}
        )

    def test_http_headers_as_param(self):
        func = mock.MagicMock(return_value={'traceparent': 'flask'})
        self.log._http_functions = {'flask': {'http_headers': func}}
        with log_module.LoggerManager(http_headers={'traceparent': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({'http_headers': {'traceparent': 'param'}}, logging.ERROR),
                {'http_headers': {'traceparent': 'param'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_headers_as_manager(self):
        func = mock.MagicMock(return_value={'traceparent': 'flask'})
        self.log._http_functions = {'flask': {'http_headers': func}}
        with log_module.LoggerManager(http_headers={'traceparent': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_headers': {'traceparent': 'manager'}, 'labels': {}}
            )
        func.assert_not_called()

    @mock.patch.dict(log_module.Logger._http_functions, {'flask': {'http_headers': mock.MagicMock(return_value={'traceparent': 'flask'})}})
    def test_http_headers_as_flask(self):
        with log_module.LoggerManager(http_headers={}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_headers': {'traceparent': 'flask'}, 'labels': {}}
            )
        self.log._http_functions['flask']['http_headers'].assert_called_once_with()

    def test_http_payload_as_param(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_payload': func}}
        with log_module.LoggerManager(http_payload={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({'http_payload': {'data': 'param'}}, logging.ERROR),
                {'http_payload': {'data': 'param'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_payload_as_manager(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_payload': func}}
        with log_module.LoggerManager(http_payload={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_payload': {'data': 'manager'}, 'labels': {}}
            )
        func.assert_not_called()

    @mock.patch.dict(log_module.Logger._http_functions, {'flask': {'http_payload': mock.MagicMock(return_value={'data': 'flask'})}})
    def test_http_payload_as_flask(self):
        with log_module.LoggerManager(http_payload={}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_payload': {'data': 'flask'}, 'labels': {}}
            )
        self.log._http_functions['flask']['http_payload'].assert_called_once_with()

    def test_traceback_with_manager(self):
        trace = ''
        try:
            raise ValueError('test')
        except ValueError:
            trace = traceback.format_exc()

        try:
            a = 1 / 0 # pylint: disable=unused-variable
        except ZeroDivisionError:
            with log_module.LoggerManager(traceback=trace):
                self.assertDictEqual(
                    self.log._get_extra_data({}, logging.ERROR),
                    {'traceback': trace, 'labels': {}}
                )

    def test_traceback(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'traceback': traceback.format_exc(), 'labels': {}}
            )

    def test_traceback_as_param(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertDictEqual(
                self.log._get_extra_data({'traceback': 'Teste'}, logging.ERROR),
                {'traceback': 'Teste', 'labels': {}}
            )

    def test_labels_with_manager(self):
        with log_module.LoggerManager(labels={'data': '123'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'labels': {'data': '123'}}
            )

    def test_labels_method(self):
        self.assertDictEqual(
            self.log._get_extra_data({'labels': {'data': '123'}}, logging.ERROR),
            {'labels': {'data': '123'}}
        )


@mock.patch.object(log_module.Logger, '_log')
class LoggerMethodsTestCase(TestCase):

    def test_critical(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.critical('critical')
        log_mock.log.assert_called_once_with(logging.CRITICAL, 'critical', extra={'labels': {}}, stacklevel=2)

    def test_debug(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.debug('debug')
        log_mock.log.assert_called_once_with(logging.DEBUG, 'debug', extra={'labels': {}}, stacklevel=2)

    def test_error(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.error('error')
        log_mock.log.assert_called_once_with(logging.ERROR, 'error', extra={'labels': {}}, stacklevel=2)

    def test_exception(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.exception('error')
        log_mock.log.assert_called_once_with(logging.ERROR, 'error', extra={'labels': {}}, stacklevel=2)

    def test_info(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.info('info')
        log_mock.log.assert_called_once_with(logging.INFO, 'info', extra={'labels': {}}, stacklevel=2)

    def test_warning(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.warning('warning')
        log_mock.log.assert_called_once_with(logging.WARNING, 'warning', extra={'labels': {}}, stacklevel=2)

    def test_deprecated(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste')
        log_mock.log.assert_called_once_with(logging.WARNING, 'DeprecationWarning: Teste', extra={'labels': {}}, stacklevel=2)

    def test_deprecated_once(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 1', show_once=True)
        log.deprecated('Teste 1', show_once=True)
        log_mock.log.assert_called_once_with(logging.WARNING, 'DeprecationWarning: Teste 1', extra={'labels': {}}, stacklevel=2)

    def test_deprecated_twice(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 2', show_once=False)
        log.deprecated('Teste 2', show_once=False)
        log_mock.log.assert_has_calls([
            mock.call(logging.WARNING, 'DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=2),
            mock.call(logging.WARNING, 'DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=2),
        ])


@mock.patch.object(log_module.Logger, '_log')
@mock.patch('everysk.core.slack.Slack')
class LoggerSlackTestCase(TestCase):

    def setUp(self) -> None:
        self.old_profile = settings.PROFILE
        self.old_slack_url = settings.SLACK_URL
        settings.PROFILE = 'PROD'
        settings.SLACK_URL = 'http://localhost'
        # This removes the module from the cache
        sys.modules.pop('unittest')

    def tearDown(self) -> None:
        settings.PROFILE = self.old_profile
        settings.SLACK_URL = self.old_slack_url
        # Put back the module
        import unittest # pylint: disable=unused-import, import-outside-toplevel

    def test_danger(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url=settings.SLACK_URL)

    def test_warning(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Warning', message='Warning message', color='warning')
        log_mock.log.assert_called_once_with(logging.WARNING, 'Slack message: Warning -> Warning message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Warning', message='Warning message', color='warning', url=settings.SLACK_URL)

    def test_success(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Success', message='Success message', color='success')
        log_mock.log.assert_called_once_with(logging.INFO, 'Slack message: Success -> Success message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Success', message='Success message', color='success', url=settings.SLACK_URL)

    def test_not_url(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        settings.SLACK_URL = None
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_not_prod(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        settings.PROFILE = 'DEV'
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_in_tests(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        import unittest # pylint: disable=unused-import, import-outside-toplevel
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_not_called()

    def test_method_url(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=2)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url='http://google.com')


class LoggerFormatterTestCase(TestCase):

    def setUp(self) -> None:
        self.formatter = log_module.Formatter()

    def test_get_default_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_dict(message='test', severity='INFO'),
            {'message': 'test', 'severity': 'INFO'}
        )

    def test_get_default_extra_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_extra_dict(
                name='everysk-log-formatter-test',
                headers={'X-Request-ID': '123'},
                payload={'data': '123'},
                traceback='',
                labels={'data': '123'}
            ),
            {
                'logName': 'everysk-log-formatter-test',
                'labels': {'data': '123'},
                'traceback': '',
                'http': {'headers': {'X-Request-ID': '123'}, 'payload': {'data': '123'}}}
        )

    def test_get_default_extra_dict_traceback(self):
        try:
            a = 1 / 0 # pylint: disable=unused-variable
        except ZeroDivisionError:
            self.assertDictEqual(
                self.formatter._get_default_extra_dict(
                    name='everysk-log-formatter-test',
                    headers={'X-Request-ID': '123'},
                    payload={'data': '123'},
                    traceback=traceback.format_exc(),
                    labels={'data': '123'}
                ),
                {
                    'logName': 'everysk-log-formatter-test',
                    'labels': {'data': '123'},
                    'traceback': traceback.format_exc(),
                    'http': {'headers': {'X-Request-ID': '123'}, 'payload': {'data': '123'}}
                }
            )

    def test_get_default_gcp_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(headers={'X-Request-ID': '123'}, filename='test.py', line=10, func_name='test_func'),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '',
                'logging.googleapis.com/trace': '',
                'logging.googleapis.com/trace_sampled': False
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_default_gcp_dict_traceparent(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(
                headers={'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'},
                filename='test.py',
                line=10,
                func_name='test_func'
            ),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '1c6c592f9e46e3fb',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_default_gcp_dict_x_cloud_trace_context(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(
                headers={'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'},
                filename='test.py',
                line=10,
                func_name='test_func'
            ),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '2048109991600514043',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_result_dict(self):
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )

        record.http_headers = {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}
        record.http_payload = {'data': '123'}
        self.assertDictEqual(
            self.formatter._get_result_dict(record),
            {
                'message': 'test',
                'severity': 'INFO',
                'labels': {},
                'logName': 'everysk-log-formatter-test',
                'traceback': '',
                'http': {
                    'headers': {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'},
                    'payload': {'data': '123'}
                },
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '1c6c592f9e46e3fb',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_result_dict_no_http(self):
        self.maxDiff = None
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )

        self.assertDictEqual(
            self.formatter._get_result_dict(record),
            {
                'message': 'test',
                'severity': 'INFO',
                'labels': {},
                'logName': 'everysk-log-formatter-test',
                'traceback': '',
                'http': {
                    'headers': {},
                    'payload': {}
                },
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '',
                'logging.googleapis.com/trace': '',
                'logging.googleapis.com/trace_sampled': False
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_format_message(self):
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )
        result = self.formatter._get_result_dict(record)
        self.assertEqual(
            self.formatter.formatMessage(record),
            json.dumps(result)
        )


@mock.patch('time.time', mock.MagicMock(return_value=1704067200.0))
class LoggerStdoutTestCase(TestCase):

    def setUp(self) -> None:
        self.log = log_module.Logger(name='everysk-log-stdout-test')

    def test_critical_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.critical('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - CRITICAL - {} - with labels\n')

    def test_critical_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.critical('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - CRITICAL - {} - without labels\n')

    def test_debug_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.debug('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - DEBUG - {} - with labels\n')

    def test_debug_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.debug('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - DEBUG - {} - without labels\n')

    def test_error_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.error('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - ERROR - {} - with labels\n')

    def test_error_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.error('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - ERROR - {} - without labels\n')

    def test_info_with_http_payload_in_bytes(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('with http payload in bytes', extra={'http_payload': b'{"key": "value"}'})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - INFO - {} - with http payload in bytes\n')
